# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from io import BytesIO
import logging
import math

import gi
from PIL import Image
from PIL import ImageChops

gi.require_version('Gdk', '4.0')
gi.require_version('Gtk', '4.0')
gi.require_version('GdkPixbuf', '2.0')
gi.require_version('Graphene', '1.0')

from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Graphene
from gi.repository import Gsk
from gi.repository import Gtk
from gi.repository.GdkPixbuf import Colorspace
from gi.repository.GdkPixbuf import Pixbuf
from gi.repository.GdkPixbuf import PixbufAnimation

from komikku.consts import MISSING_IMG_RESOURCE_PATH
from komikku.utils import get_image_info

logger = logging.getLogger('komikku')

TEXTURES_CHUNK_MAX_HEIGHT = 30000
ZOOM_FACTOR_DOUBLE_TAP = 2.5
ZOOM_FACTOR_MAX = 20
ZOOM_FACTOR_SCROLL_WHEEL = 1.3


def chunk_pixbuf(pixbuf, chunk_height):
    """Chunk a long vertical GdkPixbuf.Pixbuf into multiple GdkPixbuf.Pixbuf"""

    width = pixbuf.get_width()
    full_height = pixbuf.get_height()

    chunks = []
    for index in range(math.ceil(full_height / chunk_height)):
        y = index * chunk_height
        height = chunk_height if y + chunk_height <= full_height else full_height - y

        chunk = Pixbuf.new(Colorspace.RGB, pixbuf.get_has_alpha(), 8, width, height)
        pixbuf.copy_area(0, y, width, height, chunk, 0, 0)
        chunks.append(chunk)

    return chunks


class KImage(Gtk.Widget, Gtk.Scrollable):
    __gtype_name__ = 'KImage'
    __gsignals__ = {
        'clicked': (GObject.SignalFlags.RUN_FIRST, None, (int, int)),
        'rendered': (GObject.SignalFlags.RUN_FIRST, None, (bool, )),
        'zoom-begin': (GObject.SignalFlags.RUN_FIRST, None, ()),
        'zoom-end': (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    def __init__(self, scaling='screen', scaling_filter='linear', crop=False, landscape_zoom=False, zoomable=False):
        super().__init__()

        self.__crop = crop
        self.__hadj = None
        self.__landscape_zoom = zoomable and landscape_zoom
        self.__rendered = False
        self.__scaling = scaling
        self.__scaling_filter = scaling_filter
        self.__vadj = None
        self.__zoom = 1
        self.__zoomable = zoomable

        self.data = None
        self.path = None

        self.textures = None
        self.textures_crop = None
        self.crop_bbox = None

        self.animation = None  # PixbufAnimation
        self.animation_iter = None
        self.animation_tick_callback_id = None

        self.hadjustment_value_changed_handler_id = None
        self.vadjustment_value_changed_handler_id = None

        self.gesture_click_timeout_id = None
        self.pointer_position = None  # current pointer position
        self.zoom_center = None  # zoom position in image
        self.zoom_gesture_begin = None
        self.zoom_scaling = None  # zoom factor at scaling

        self.set_overflow(Gtk.Overflow.HIDDEN)

        if self.__zoomable:
            # Controller to track pointer motion: used to know current cursor position
            self.controller_motion = Gtk.EventControllerMotion.new()
            self.add_controller(self.controller_motion)
            self.controller_motion.connect('motion', self.on_pointer_motion)

            # Controller to zoom with Ctrl + mouse wheel or 2-fingers touchpad gesture
            self.controller_scroll = Gtk.EventControllerScroll.new(Gtk.EventControllerScrollFlags.VERTICAL)
            self.controller_scroll.set_propagation_phase(Gtk.PropagationPhase.CAPTURE)
            self.add_controller(self.controller_scroll)
            self.controller_scroll.connect('scroll', self.on_scroll)

            # Gesture click controller: double-tap zoom
            self.gesture_click = Gtk.GestureClick.new()
            self.gesture_click.set_propagation_phase(Gtk.PropagationPhase.CAPTURE)
            self.gesture_click.set_button(1)
            self.gesture_click.connect('released', self.on_gesture_click_released)
            self.add_controller(self.gesture_click)

            # Gesture zoom controller (2-fingers touchpad/touchscreen gesture)
            self.gesture_zoom = Gtk.GestureZoom.new()
            self.gesture_zoom.set_propagation_phase(Gtk.PropagationPhase.CAPTURE)
            self.gesture_zoom.connect('begin', self.on_gesture_zoom_begin)
            self.gesture_zoom.connect('end', self.on_gesture_zoom_end)
            self.gesture_zoom.connect('scale-changed', self.on_gesture_zoom_scale_changed)
            self.add_controller(self.gesture_zoom)

    @property
    def borders(self):
        """ Width of vertical (top, bottom) and horizontal (left, right) bars """
        if self.widget_width > self.image_displayed_width:
            hborder = (self.widget_width - self.image_displayed_width) / 2
        else:
            hborder = 0

        if self.widget_height > self.image_displayed_height:
            vborder = (self.widget_height - self.image_displayed_height) / 2
        else:
            vborder = 0

        return (hborder, vborder)

    @GObject.Property(type=bool, default=False)
    def crop(self):
        return self.__crop and not self.animation

    @crop.setter
    def crop(self, value):
        if self.__crop == value or self.animation:
            return

        self.__crop = value
        self.queue_resize()

    @GObject.Property(type=Gtk.Adjustment)
    def hadjustment(self):
        return self.__hadj or Gtk.Adjustment()

    @hadjustment.setter
    def hadjustment(self, adj):
        if not adj:
            return

        self.__hadj = adj
        self.hadjustment_value_changed_handler_id = adj.connect('value-changed', lambda _adj: self.queue_draw())
        self.configure_adjustments()

    @GObject.Property(type=Gtk.ScrollablePolicy, default=Gtk.ScrollablePolicy.MINIMUM)
    def hscroll_policy(self):
        return Gtk.ScrollablePolicy.MINIMUM

    @property
    def image_height(self):
        """ Image original height """
        if self.animation:
            return self.animation.get_height()

        if self.crop and self.crop_bbox:
            return self.crop_bbox[3] - self.crop_bbox[1]

        return sum(texture.get_height() for texture in self.textures) if self.textures else 0

    @property
    def image_width(self):
        """ Image original width """
        if self.animation:
            return self.animation.get_width()

        if self.crop and self.crop_bbox:
            return self.crop_bbox[2] - self.crop_bbox[0]

        return self.textures[0].get_width() if self.textures else 0

    @property
    def image_displayed_height(self):
        """ Image height with current zoom factor """
        return int(self.image_height * self.zoom)

    @property
    def image_displayed_width(self):
        """ Image width with current zoom factor """
        return int(self.image_width * self.zoom)

    @GObject.Property(type=bool, default=False)
    def landscape_zoom(self):
        return self.__landscape_zoom

    @landscape_zoom.setter
    def landscape_zoom(self, value):
        if self.__landscape_zoom == value:
            return

        self.__landscape_zoom = value
        self.queue_resize()

    @property
    def max_hadjustment_value(self):
        return max(self.image_displayed_width - self.widget_width, 0)

    @property
    def max_vadjustment_value(self):
        return max(self.image_displayed_height - self.widget_height, 0)

    @property
    def obscured(self):
        parent = self.get_parent()
        if isinstance(parent, Gtk.ScrolledWindow):
            return parent.get_parent().obscured

        return parent.obscured

    @property
    def ratio(self):
        return self.image_width / self.image_height

    @GObject.Property(type=str, default='screen')
    def scaling(self):
        """ Type of scaling:
        - adapt to screen (best-fit)
        - adapt to width
        - adapt to height
        - original size
        """
        return self.__scaling

    @scaling.setter
    def scaling(self, value):
        if self.__scaling == value:
            return

        self.__scaling = value
        self.queue_resize()

    @GObject.Property(type=str, default='linear')
    def scaling_filter(self):
        """ Scaling filters:
        - linear
        - trilinear
        """
        return self.__scaling_filter

    @scaling_filter.setter
    def scaling_filter(self, value):
        if self.__scaling_filter == value:
            return

        self.__scaling_filter = value
        self.queue_resize()

    @property
    def scaling_size(self):
        """ Image size at defined scaling """
        scaling = self.scaling
        if scaling != 'original':
            if self.landscape_zoom and scaling == 'screen' and self.image_width > self.image_height:
                # When page is landscape and scaling is 'screen', scale page to fit height
                scaling = 'height'

            max_width = self.widget_width
            max_height = self.widget_height

            adapt_to_width_height = max_width * self.image_height / self.image_width
            adapt_to_height_width = max_height * self.image_width / self.image_height

            if scaling == 'width' or (scaling == 'screen' and adapt_to_width_height <= max_height):
                # Adapt image to width
                width = max_width
                height = adapt_to_width_height
            elif scaling == 'height' or (scaling == 'screen' and adapt_to_height_width <= max_width):
                # Adapt image to height
                width = adapt_to_height_width
                height = max_height
        else:
            width = self.image_width
            height = self.image_height

        return (width, height)

    @property
    def scrollable(self):
        return isinstance(self.get_parent(), Gtk.ScrolledWindow)

    @GObject.Property(type=Gtk.Adjustment)
    def vadjustment(self):
        return self.__vadj or Gtk.Adjustment()

    @vadjustment.setter
    def vadjustment(self, adj):
        if not adj:
            return

        self.__vadj = adj
        self.vadjustment_value_changed_handler_id = adj.connect('value-changed', lambda _adj: self.queue_draw())
        self.configure_adjustments()

    @GObject.Property(type=Gtk.ScrollablePolicy, default=Gtk.ScrollablePolicy.MINIMUM)
    def vscroll_policy(self):
        return Gtk.ScrollablePolicy.MINIMUM

    @property
    def widget_height(self):
        return self.get_height()

    @property
    def widget_width(self):
        return self.get_width()

    @GObject.Property(type=float)
    def zoom(self):
        """ Displayed zoom level """
        return self.__zoom

    @zoom.setter
    def zoom(self, value):
        if self.__zoom == value:
            return
        self.__zoom = value
        self.queue_resize()

    @property
    def zoomable(self):
        return self.__zoomable

    def __animation_tick_callback(self, image, clock):
        if self.animation is not None and self.animation_iter is None:
            return GLib.SOURCE_REMOVE

        # Do not animate if not visible (obscured)
        if self.obscured or not self.get_mapped():
            return GLib.SOURCE_CONTINUE

        # Check if it's time to show the next frame
        if self.animation_iter.advance(None):
            self.queue_draw()

        return GLib.SOURCE_CONTINUE

    def __compute_borders_crop_bbox(self):
        threshold = 225

        if self.path is None and self.data is None:
            return None

        def lookup(x):
            return 255 if x > threshold else 0

        try:
            with Image.open(self.path or BytesIO(self.data)) as im:
                with im.convert('L') as im_bw:
                    im_lookup = im_bw.point(lookup, mode='1')
        except Exception as exc:
            logger.error('Failed to compute image white borders bbox', exc_info=exc)
            return None

        with Image.new(im_lookup.mode, im_lookup.size, 255) as im_bg:
            return ImageChops.difference(im_lookup, im_bg).getbbox()

    def cancel_deceleration(self):
        if self.scrollable:
            self.get_parent().set_kinetic_scrolling(False)
            self.get_parent().set_kinetic_scrolling(True)

    def configure_adjustments(self):
        self.hadjustment.configure(
            # value
            max(min(self.hadjustment.props.value, self.max_hadjustment_value), 0),
            # lower value
            0,
            # upper value
            self.image_displayed_width,
            # step increment
            self.widget_width * 0.1,
            # page increment
            self.widget_width * 0.9,
            # page size
            min(self.widget_width, self.image_displayed_width)
        )

        self.vadjustment.configure(
            max(min(self.vadjustment.props.value, self.max_vadjustment_value), 0),
            0,
            self.image_displayed_height,
            self.widget_height * 0.1,
            self.widget_height * 0.9,
            min(self.widget_height, self.image_displayed_height)
        )

    def crop_borders(self):
        """ Crop white borders """
        bbox = self.crop_bbox
        textures_width = self.textures[0].get_width()
        textures_height = sum(texture.get_height() for texture in self.textures)

        # Crop is possible if computed bbox is included in textures
        if bbox and (bbox[2] - bbox[0] < textures_width or bbox[3] - bbox[1] < textures_height):
            try:
                with Image.open(self.path or BytesIO(self.data)) as im:
                    with im.convert('RGB') as im_rgb:
                        with im_rgb.crop(bbox) as im_crop:
                            with BytesIO() as io_buffer:
                                # Use the very fast TIFF format (Pillow uses libtiff)
                                im_crop.save(io_buffer, 'tiff')

                                if bbox[3] - bbox[1] > TEXTURES_CHUNK_MAX_HEIGHT:
                                    stream = Gio.MemoryInputStream.new_from_data(io_buffer.getvalue(), None)

                                    textures = []
                                    for pix in chunk_pixbuf(Pixbuf.new_from_stream(stream), TEXTURES_CHUNK_MAX_HEIGHT):
                                        textures.append(Gdk.Texture.new_for_pixbuf(pix))

                                    stream.close()

                                    return textures
                                else:
                                    return [Gdk.Texture.new_from_bytes(GLib.Bytes.new(io_buffer.getvalue()))]
            except Exception as exc:
                logger.error('Failed to crop image white borders', exc_info=exc)

        return self.textures

    def dispose(self):
        if self.hadjustment_value_changed_handler_id:
            self.hadjustment.disconnect(self.hadjustment_value_changed_handler_id)
        if self.vadjustment_value_changed_handler_id:
            self.vadjustment.disconnect(self.vadjustment_value_changed_handler_id)

        if self.zoomable:
            self.controller_motion.disconnect_by_func(self.on_pointer_motion)
            self.controller_scroll.disconnect_by_func(self.on_scroll)
            self.gesture_click.disconnect_by_func(self.on_gesture_click_released)
            self.gesture_zoom.disconnect_by_func(self.on_gesture_zoom_begin)
            self.gesture_zoom.disconnect_by_func(self.on_gesture_zoom_end)
            self.gesture_zoom.disconnect_by_func(self.on_gesture_zoom_scale_changed)

        if self.animation_tick_callback_id:
            self.remove_tick_callback(self.animation_tick_callback_id)

        self.data = None
        self.textures = None
        self.textures_crop = None
        self.animation_iter = None
        self.animation = None

    def do_measure(self, orientation, for_size):
        if orientation == Gtk.Orientation.HORIZONTAL:
            return 0, int(for_size * self.ratio) if for_size != -1 else -1, -1, -1

        return 0, int(for_size / self.ratio) if for_size != -1 else -1, -1, -1

    def do_size_allocate(self, w, h, b):
        if self.crop and self.crop_bbox is None:
            self.crop_bbox = self.__compute_borders_crop_bbox()

        if self.zoom_scaling is None or self.zoom == self.zoom_scaling:
            self.zoom_scaling = self.scaling_size[1] / self.image_height
            self.set_zoom()
        else:
            self.configure_adjustments()

    def do_snapshot(self, snapshot):
        if self.animation_iter:
            # Get next frame (animated GIF)
            self.textures = [Gdk.Texture.new_for_pixbuf(self.animation_iter.get_pixbuf())]

        elif self.crop and self.textures_crop is None and self.textures:
            # Crop white borders
            self.textures_crop = self.crop_borders()

        self.configure_adjustments()

        snapshot.save()

        width = self.image_displayed_width
        height = self.image_displayed_height

        if self.scrollable:
            x = -(self.hadjustment.props.value - (self.hadjustment.props.upper - width) / 2)
            snapshot.translate(Graphene.Point().init(int(x), 0))
            y = -(self.vadjustment.props.value - (self.vadjustment.props.upper - height) / 2)
            snapshot.translate(Graphene.Point().init(0, int(y)))

        # Center in widget
        snapshot.translate(
            Graphene.Point().init(
                max((self.widget_width - width) // 2, 0),
                max((self.widget_height - height) // 2, 0),
            )
        )

        # Append textures
        scale_factor = self.get_native().get_surface().get_scale()
        if scale_factor != 1:
            snapshot.scale(1 / scale_factor, 1 / scale_factor)

        filter = Gsk.ScalingFilter.LINEAR if self.scaling_filter == 'linear' else Gsk.ScalingFilter.TRILINEAR
        rect = Graphene.Rect().alloc()
        textures = self.textures_crop if self.crop else self.textures
        y = 0
        for texture in textures:
            h = int(texture.get_height() * self.zoom) * scale_factor
            rect.init(0, y, int(texture.get_width() * self.zoom) * scale_factor, h)
            snapshot.append_scaled_texture(texture, filter, rect)
            y += h

        snapshot.restore()

        self.emit('rendered', self.__rendered)
        if not self.__rendered:
            self.__rendered = True

    def load(self, path=None, data=None, callback=None, static_animation=False):
        info, mime_type = get_image_info(path or data)
        if info is None:
            self.load_missing(callback=callback, error='corrupted-or-unsupported-format', mime_type=mime_type)
            return

        try:
            if path:
                with open(path, 'rb') as fp:
                    stream = Gio.MemoryInputStream.new_from_data(fp.read(), None)
            elif data:
                stream = Gio.MemoryInputStream.new_from_data(data, None)

            if not info['is_animated']:
                Pixbuf.new_from_stream_async(stream, None, self.load_ready, callback, info, mime_type)
            else:
                PixbufAnimation.new_from_stream_async(stream, None, self.load_ready, callback, info, mime_type)

        except Exception as exc:
            logger.warning('Failed to create textures: corrupted file or unsupported image format', exc_info=exc)
            self.load_missing(callback=callback, error='corrupted-or-unsupported-format', mime_type=mime_type)
            return

        self.path = path
        self.data = data

    def load_missing(self, callback=None, error=None, mime_type=None):
        self.path = MISSING_IMG_RESOURCE_PATH

        # Reset instance variables and controllers
        self.dispose()
        self.__scaling = 'screen'
        self.__scaling_filter = 'linear'
        self.__crop = False
        self.__landscape_zoom = False
        self.__zoomable = False

        try:
            self.textures = [Gdk.Texture.new_from_resource(self.path)]

        except Exception as exc:
            logger.warning('Failed to create textures: corrupted file or unsupported image format', exc_info=exc)
            return

        if callback:
            callback(self, error=error, mime_type=mime_type)

    def load_ready(self, stream, result, callback, info, mime_type):
        stream.close()

        try:
            if not info['is_animated']:
                pixbuf = Pixbuf.new_from_stream_finish(result)

                self.textures = []
                if info['height'] < TEXTURES_CHUNK_MAX_HEIGHT:
                    self.textures.append(Gdk.Texture.new_for_pixbuf(pixbuf))
                else:
                    # Chunk a long vertical pixbuf into several textures
                    for cpixbuf in chunk_pixbuf(pixbuf, TEXTURES_CHUNK_MAX_HEIGHT):
                        self.textures.append(Gdk.Texture.new_for_pixbuf(cpixbuf))

            else:
                self.animation = PixbufAnimation.new_from_stream_finish(result)
                self.animation_iter = self.animation.get_iter(None)
                self.animation_tick_callback_id = self.add_tick_callback(self.__animation_tick_callback)

        except Exception as exc:
            logger.warning('Failed to create textures: corrupted file or unsupported image format', exc_info=exc)
            self.load_missing(callback=callback, error='corrupted-or-unsupported-format', mime_type=mime_type)
            return

        callback(self)

    def on_gesture_click_released(self, _gesture, n_press, x, y):
        def emit_clicked(x, y):
            GLib.source_remove(self.gesture_click_timeout_id)
            self.gesture_click_timeout_id = None
            self.emit('clicked', x, y)

        if n_press == 1 and self.gesture_click_timeout_id is None and self.zoom == self.zoom_scaling:
            # Schedule single click event to be able to detect double click
            dbl_click_time = Gtk.Settings.get_default().get_property('gtk-double-click-time')
            self.gesture_click_timeout_id = GLib.timeout_add(dbl_click_time, emit_clicked, x, y)

        elif n_press == 2:
            # Remove scheduled single click event
            if self.gesture_click_timeout_id:
                GLib.source_remove(self.gesture_click_timeout_id)
                self.gesture_click_timeout_id = None

            if self.zoom == self.zoom_scaling:
                self.set_zoom(self.zoom * ZOOM_FACTOR_DOUBLE_TAP, (x, y))
            else:
                self.set_zoom(self.zoom_scaling)

    def on_gesture_zoom_begin(self, _gesture, _sequence):
        self.cancel_deceleration()

        _active, x, y = self.gesture_zoom.get_bounding_box_center()
        self.zoom_center = (x, y)
        self.zoom_gesture_begin = self.zoom

        self.gesture_zoom.set_state(Gtk.EventSequenceState.CLAIMED)

    def on_gesture_zoom_end(self, _gesture, _sequence):
        self.gesture_zoom.set_state(Gtk.EventSequenceState.CLAIMED)

    def on_gesture_zoom_scale_changed(self, _gesture, scale):
        self.set_zoom(self.zoom_gesture_begin * scale, self.zoom_center)

        if self.gesture_zoom.get_device().get_source() == Gdk.InputSource.TOUCHSCREEN and self.zoom_center:
            # Move image to follow zoom position on touchscreen
            _active, x, y = self.gesture_zoom.get_bounding_box_center()
            self.hadjustment.set_value(self.hadjustment.get_value() + self.zoom_center[0] - x)
            self.vadjustment.set_value(self.vadjustment.get_value() + self.zoom_center[1] - y)
            self.zoom_center = (x, y)

        self.gesture_zoom.set_state(Gtk.EventSequenceState.CLAIMED)

    def on_pointer_motion(self, _controller, x, y):
        self.pointer_position = (x, y)

    def on_scroll(self, _controller, _dx, dy):
        """ <Ctrl>+Scroll zooming """
        modifiers = Gtk.accelerator_get_default_mod_mask()
        state = self.controller_scroll.get_current_event_state()
        if state & modifiers == Gdk.ModifierType.CONTROL_MASK:
            factor = math.exp(-dy * math.log(ZOOM_FACTOR_SCROLL_WHEEL))
            self.set_zoom(self.zoom * factor, self.pointer_position)
            return Gdk.EVENT_STOP

        return Gdk.EVENT_PROPAGATE

    def set_allow_zooming(self, allow):
        if not self.zoomable:
            return

        self.controller_scroll.set_propagation_phase(Gtk.PropagationPhase.CAPTURE if allow else Gtk.PropagationPhase.NONE)
        self.gesture_zoom.set_propagation_phase(Gtk.PropagationPhase.CAPTURE if allow else Gtk.PropagationPhase.NONE)

    def set_zoom(self, zoom=None, center=None):
        """ Set zoom level for given position """

        if zoom is None:
            zoom = self.zoom_scaling
        else:
            zoom = min(max(zoom, self.zoom_scaling), ZOOM_FACTOR_MAX)

            if zoom != self.zoom_scaling and self.zoom == self.zoom_scaling:
                self.emit('zoom-begin')
            elif zoom == self.zoom_scaling and self.zoom != self.zoom_scaling:
                self.emit('zoom-end')

        if center:
            borders = self.borders
            zoom_ratio = self.zoom / zoom

        self.zoom = zoom

        self.configure_adjustments()

        if center:
            hadjustment_value = self.hadjustment.get_value()
            vadjustment_value = self.vadjustment.get_value()

            x = center[0]
            y = center[1]

            value = max((x + hadjustment_value - borders[0]) / zoom_ratio - x, 0)
            self.hadjustment.set_value(value)

            value = max((y + vadjustment_value - borders[1]) / zoom_ratio - y, 0)
            self.vadjustment.set_value(value)

    def set_zoom_by_key(self, keyval, reading_mode):
        # Compute center depending on reading mode
        if reading_mode == 'left-to-right':
            # Top left
            center = (0, 0)

        elif reading_mode == 'vertical':
            # Center top
            center = (self.widget_width // 2, 0)

        else:
            # Top right
            center = (self.widget_width, 0)

        if keyval in (Gdk.KEY_plus, Gdk.KEY_KP_Add):
            # Zoom in
            self.set_zoom(self.zoom * ZOOM_FACTOR_SCROLL_WHEEL, center)

        elif keyval in (Gdk.KEY_minus, Gdk.KEY_KP_Subtract):
            # Zoom out
            self.set_zoom(self.zoom / ZOOM_FACTOR_SCROLL_WHEEL, center)

        elif keyval in (Gdk.KEY_0, Gdk.KEY_KP_0):
            # Reset
            self.set_zoom()



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from abc import ABC
from abc import abstractmethod
from functools import wraps
from gettext import gettext as _
import json
import logging
import os
import threading
import time

import jwt

from gi.repository import GObject

from komikku.consts import LOGO_SIZE
from komikku.models import create_db_connection
from komikku.models import Manga
from komikku.models import Settings
from komikku.utils import BaseServer
from komikku.utils import get_cached_logos_dir
from komikku.utils import log_error_traceback
from komikku.trackers.utils import get_trackers_list

logger = logging.getLogger(__name__)


def authenticated(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        tracker = args[0]
        granted, access_token_valid = tracker.is_granted()

        if not granted:
            return None

        if access_token_valid:
            return func(*args, **kwargs, access_token=tracker.data['access_token'])

        # Access token is expired, try to refresh it
        if new_access_token := tracker.refresh_access_token():
            logger.info('%s tracker access token has been successfully refreshed', tracker.name)
            return func(*args, **kwargs, access_token=new_access_token)

        # Tracker don't support refresh or refresh has failed
        return None

    return wrapper


class Tracker(BaseServer, ABC):
    id: str
    name: str
    base_url: str

    authorize_url: str = None
    app_redirect_url: str = 'https://komikku.info/'
    manga_url: str = None
    logo_url: str = None

    headers_images = {}

    INTERNAL_STATUSES = {
        'reading': _('Reading'),
        'completed': _('Completed'),
        'on_hold': _('On Hold'),
        'dropped': _('Dropped'),
        'plan_to_read': _('Plan to Read'),
        'rereading': _('Rereading'),
    }
    STATUSES_MAPPING: dict = None

    @property
    def data(self):
        """Tracker data saved in settings"""
        return Settings.get_default().trackers.get(self.id)

    @data.setter
    def data(self, data):
        """Save tracker data (access token, refresh token,...) in settings"""
        trackers = Settings.get_default().trackers

        trackers[self.id] = data

        Settings.get_default().trackers = trackers

    @property
    def logo_path(self):
        path = os.path.join(get_cached_logos_dir(), 'trackers', f'{self.id}.png')
        if not os.path.exists(path):
            return None

        return path

    def convert_internal_status(self, status):
        """Returns corresponding tracker status for an internal status"""
        for tracker_status, internal_status in self.STATUSES_MAPPING.items():
            if internal_status == status:
                return tracker_status

    def get_manga_data(self, id):
        data = self.get_user_manga_data(id)
        if data is None:
            data = self.get_tracker_manga_data(id)
            if data:
                # Add user's status fields that doesn't exist in tracker manga data
                data['chapters_progress'] = 0
                data['score'] = 0
                data['status'] = 'reading'

        return data

    @abstractmethod
    def get_manga_url(self, id):
        """Returns manga URL"""

    def get_status_from_index(self, index):
        return list(self.STATUSES_MAPPING.values())[index]

    def get_status_index(self, internal_status):
        for index, status in enumerate(self.STATUSES_MAPPING.values()):
            if status == internal_status:
                return index

    @abstractmethod
    def get_tracker_manga_data(self, id):
        """Retrieves tracker manga info"""

    @abstractmethod
    def get_user_manga_data(self, id):
        """Retrieves user manga info (progress, score, status)"""

    @abstractmethod
    def get_user_score_format(self, format):
        """Returns user score format (min, max, step, raw factor)"""

    def is_granted(self):
        """
        :return: is granted, is access token valid (not expired)
        :rtype: (bool, bool)
        """
        data = self.data
        if data is None:
            return False, False

        if not data.get('access_token'):
            return False, False

        # Get expires timestamp in token payload
        payload = jwt.decode(data['access_token'], options={'verify_signature': False})
        expiration_time = payload.get('exp')
        # Else, get it in data if available
        if expiration_time is None:
            expiration_time = data.get('exp')

        if expiration_time is None:
            # No expiration time, token is assumed to be valid
            return True, True

        return True, time.time() <= expiration_time - 86400

    @abstractmethod
    def refresh_access_token(self):
        """
        Refreshes the access token

        :return: The new access token
        :rtype: string
        """

    @abstractmethod
    def request_access_token(self):
        """Requests an access token"""

    def save_logo(self):
        return self.save_image(
            self.logo_url, os.path.join(get_cached_logos_dir(), 'trackers'), self.id,
            LOGO_SIZE, LOGO_SIZE, keep_aspect_ratio=False, format='PNG'
        )

    @abstractmethod
    def search(self, term):
        """Search a manga"""

    @abstractmethod
    def update_user_manga_data(self, id, data):
        """"Updates user manga data"""


class Trackers(GObject.GObject):
    __gsignals__ = {
        'manga-tracker-synced': (GObject.SignalFlags.RUN_FIRST, None, (GObject.TYPE_PYOBJECT, )),
    }

    def __init__(self, window):
        super().__init__()
        self.window = window

        self.trackers = {}
        for info in get_trackers_list():
            tracker = getattr(info['module'], info['class_name'])()
            self.trackers[tracker.id] = tracker

    def sync(self):
        def run():
            db_conn = create_db_connection()

            for id, tracker in self.trackers.items():
                granted, access_token_valid = tracker.is_granted()
                if not granted or not access_token_valid:
                    continue

                query = f"SELECT id, tracking -> '{id}' AS data FROM mangas WHERE tracking -> '$.{id}._synced' = 'false'"

                for row in db_conn.execute(query).fetchall():
                    data = json.loads(row['data'])
                    try:
                        res = tracker.update_user_manga_data(data['id'], {
                            'score': data['score'],
                            'chapters_progress': data['chapters_progress'],
                            'status': data['status'],
                        })
                    except Exception as e:
                        res = False
                        log_error_traceback(e)
                        logging.warning(f'Failed to sync tracker {id}: ID={data["id"]} name={data["name"]}')

                    if res:
                        manga = Manga.get(row['id'], db_conn=db_conn)
                        manga.tracking[id]['_synced'] = True
                        manga.update({
                            'tracking': manga.tracking,
                        })

                        self.emit('manga-tracker-synced', manga)
                    else:
                        logging.warning(f'Failed to sync tracker {id}: ID={data["id"]} name={data["name"]}')

        if not Settings.get_default().tracking:
            return

        thread = threading.Thread(target=run)
        thread.daemon = True
        thread.start()



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from bs4 import BeautifulSoup
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.utils import get_buffer_mime_type
from komikku.utils import is_number

# An alternative site exists (Madara)
# https://tcbscans-manga.com


class Tcbscans(Server):
    id = 'tcbscans'
    name = 'TCB Scans'
    lang = 'en'

    base_url = 'https://tcbonepiecechapters.com'
    logo_url = base_url + '/files/favicon-32x32.png'
    most_populars_url = base_url + '/projects'
    manga_url = base_url + '/mangas/{0}'
    chapter_url = base_url + '/chapters/{0}'
    image_url = 'https://cdn.onepiecechapters.com/file/CDN-M-A-N/{0}'

    def __init__(self):
        self.session = requests.Session()
        self.session.headers = {
            'User-Agent': USER_AGENT,
        }

    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Manga slug is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[self.name, ],
            genres=[],
            status=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
            cover=None,
        ))

        data['name'] = soup.find('h1').text.strip()
        data['cover'] = soup.select_one('.grid .bg-card > .flex > img').get('src')

        # Details
        data['synopsis'] = soup.select_one('p.leading-6').text.strip()

        # Chapters
        for a_element in reversed(soup.select('a.block')):
            slug = '/'.join(a_element.get('href').split('/')[-2:])
            num = a_element.select_one('div:first-child').text.strip().split()[-1]
            title = a_element.select_one('div:last-child').text.strip()

            data['chapters'].append(dict(
                slug=slug,
                title=f'Chapter {num} - {title}' if title else f'Chapter {num}',
                num=num if is_number(num) else None,
                date=None,
            ))

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Currently, only pages are expected.
        """
        r = self.session_get(self.chapter_url.format(chapter_slug))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = dict(
            pages=[],
        )
        # Two different selectors are required
        # Some chapters page requests (last chapter of One piece) are redirected (https://jkstudyupdates.in/latestchapter.html)
        for img_element in soup.select('picture img, .image-container img'):
            data['pages'].append(dict(
                slug=img_element.get('src').split('/')[-1],
                image=None,
            ))

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(
            self.image_url.format(page['slug']),
            headers={
                'Referer': self.chapter_url.format(chapter_slug),
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=page['slug'],
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_most_populars(self):
        """
        Returns manga list
        """
        r = self.session_get(self.most_populars_url)
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for a_element in soup.select('.grid .bg-card .flex .relative a'):
            results.append(dict(
                slug='/'.join(a_element.get('href').split('/')[-2:]),
                name=a_element.img.get('alt'),
                cover=a_element.img.get('src'),
            ))

        return results

    def search(self, term):
        results = []
        for item in self.get_most_populars():
            if term and term.lower() in item['name'].lower():
                results.append(item)

        return results



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import json

from bs4 import BeautifulSoup
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.multi.heancms import HeanCMS
from komikku.servers.multi.genkan import GenkanInitial
from komikku.servers.multi.madara import Madara
from komikku.servers.utils import convert_date_string
from komikku.servers.utils import parse_nextjs_hydration
from komikku.utils import get_buffer_mime_type


class Reaperscans(HeanCMS):
    id = 'reaperscans'
    name = 'Reaper Scans'
    lang = 'en'
    status = 'disabled'

    base_url = 'https://reaperscans.com'
    logo_url = base_url + '/wetried_only.png'
    api_url = 'https://api.reaperscans.com'
    api_version = 2
    api_chapters_url = api_url + '/chapters/{0}'

    name_css_path = 'h1'
    cover_css_path = 'img[width="316"]'
    authors_css_path = 'div.flex:-soup-contains("Author") > span:last-child'
    synopsis_css_path = 'div.text-muted-foreground > div:nth-child(1)'

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Pages URLs are available in a <script> element
        """
        r = self.session_get(
            self.chapter_url.format(manga_slug, chapter_slug),
            headers={
                'Referer': self.manga_url.format(manga_slug),
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        if info := parse_nextjs_hydration(soup, 'API_Response'):
            data = dict(
                pages=[],
            )
            index = 0
            for block in info[1][3]['children'][0]:
                pages = block[3]['children']
                if isinstance(pages[0][0], list):
                    pages = pages[0]

                for page in pages:
                    data['pages'].append(dict(
                        slug=None,
                        image=page[3]['src'],
                        index=index + 1,
                    ))
                    index += 1

            return data

        return None


class Reaperscans_ar(Madara):
    id = 'reaperscans_ar'
    name = 'ريبر العربي'
    lang = 'ar'
    status = 'disabled'

    series_name = 'series'
    date_format = '%Y, %d %B'

    base_url = 'https://reaperscansar.com'
    chapters_url = base_url + '/series/{0}/ajax/chapters/'


class Reaperscans_fr(Madara):
    id = 'reaperscans_fr'
    name = 'ReaperScansFR (GS)'
    lang = 'fr'
    status = 'disabled'

    has_cf = True

    series_name = 'serie'
    date_format = '%d/%m/%Y'

    base_url = 'https://reaper-scans.fr'  # Madara => Keyoapp

    details_scanlators_selector = '.post-content_item:-soup-contains("Team") .summary-content'


class Reaperscans_id(Madara):
    id = 'reaperscans_id'
    name = 'Reaper Scans'
    lang = 'id'
    status = 'disabled'

    series_name = 'series'

    base_url = 'https://reaperscans.id'


class Reaperscans_pt(Server):
    id = 'reaperscans_pt'
    name = 'Reaper Scans'
    lang = 'pt'
    status = 'disabled'  # Switch to HeanCMS (2023-??), a new server has been added with correct language (pt-BR)

    api_base_url = 'https://api.reaperscans.net'
    api_search_url = api_base_url + '/series/search'
    api_most_populars_url = api_base_url + '/series/querysearch'
    api_chapter_url = api_base_url + '/series/chapter/{}'

    base_url = 'https://reaperscans.net'
    manga_url = base_url + '/series/{0}'
    chapter_url = base_url + '/series/{0}/{1}'

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update({'user-agent': USER_AGENT})

    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Slug is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],  # Not available
            genres=[],
            status=None,    # Not available
            cover=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
        ))

        data['name'] = soup.find('h1').text.strip()
        data['cover'] = soup.find('img', class_='rounded-thumb').get('src')

        # Details
        data['genres'] = [span.text.strip() for span in soup.find('div', class_='tags-container').find_all('span', class_='tag')]
        data['status'] = 'ongoing'

        container_element = soup.find('div', class_='useful-container')
        for author in container_element.select_one('p:-soup-contains("Autor") strong').text.strip().split(','):
            data['authors'].append(author.strip())

        # Synopsis
        data['synopsis'] = soup.find('div', class_='description-container').text.strip()

        # Chapters
        for a_element in reversed(soup.select('#simple-tabpanel-0 ul > a')):
            data['chapters'].append(dict(
                slug=a_element.get('href').split('/')[-1],
                title=a_element.select_one('.MuiTypography-body1').text.strip(),
                date=convert_date_string(a_element.select_one('.MuiTypography-body2').text.strip(), format='%d/%m/%Y'),
            ))

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content + API

        Currently, only pages are expected.
        """
        r = self.session_get(self.chapter_url.format(manga_slug, chapter_slug))
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        id_ = None
        for script_element in soup.find_all('script'):
            if script_element.get('id') != '__NEXT_DATA__':
                continue
            data = json.loads(script_element.string)
            id_ = data['props']['pageProps']['data']['id']
            break

        if id_ is None:
            return None

        r = self.session_get(self.api_chapter_url.format(id_))
        if r.status_code != 200:
            return None

        data = dict(
            pages=[],
        )
        for image in r.json()['content']['images']:
            data['pages'].append(dict(
                slug=None,
                image=image,
            ))

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(self.api_base_url + '/' + page['image'])
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=page['image'].split('/')[-1],
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_latest_updates(self):
        return self.search('', orderby='latest')

    def get_most_populars(self):
        return self.search('', orderby='populars')

    def search(self, term, orderby=None):
        if orderby:
            r = self.session_post(
                self.api_most_populars_url,
                params=dict(
                    order='desc',
                    order_by='total_views' if orderby == 'populars' else 'recently_added',
                    series_type='Comic',
                ),
                headers={
                    'content-type': 'application/json',
                }
            )
        else:
            r = self.session_post(
                self.api_search_url,
                params=dict(
                    term=term,
                ),
                headers={
                    'content-type': 'application/json',
                }
            )
        if r.status_code != 200:
            return None

        items = r.json()
        if orderby:
            items = items['data']

        results = []
        for item in items:
            if item['series_type'] not in ('Comic',):
                continue

            results.append(dict(
                slug=item['series_slug'],
                name=item['title'],
            ))

        return results


class Reaperscans_pt_br(HeanCMS):
    id = 'reaperscans_pt_br'
    name = 'Yugen Scans (Reaper Scans)'
    lang = 'pt_BR'
    status = 'disabled'  # 08/2024: move to https://yugenweb.com

    base_url = 'https://yugenmangas.lat'
    api_url = 'https://api.yugenmangas.net'

    cover_css_path = 'div div div.container.px-5.text-gray-50 div.grid.grid-cols-12.pt-3.gap-x-3 div.col-span-12.relative.flex.justify-center.w-full div.flex.flex-col.items-center.justify-center.gap-y-2.w-full img'
    authors_css_path = 'div div.container.px-5.text-gray-50 div.grid.grid-cols-12.pt-3.gap-x-3 div.col-span-12.flex.flex-col.gap-y-3 div div.flex.flex-col.gap-y-2 p:nth-child(3) strong'
    synopsis_css_path = 'div div.container.px-5.text-gray-50 div.grid.grid-cols-12.pt-3.gap-x-3 div.col-span-12.flex.flex-col.gap-y-3 div.bg-gray-800.text-gray-50.rounded-xl.p-5'


class Reaperscans_tr(Madara):
    id = 'reaperscans_tr'
    name = 'Reaper Scans'
    lang = 'tr'
    status = 'disabled'  # 08/2024: move to https://yugenweb.com

    series_name = 'seri'

    base_url = 'https://reaperscanstr.com'


class Reaperscans__old(GenkanInitial):
    id = 'reaperscans__old'
    name = 'Reaper Scans'
    lang = 'en'
    status = 'disabled'

    # Use Cloudflare
    # Search is partially broken -> inherit from GenkanInitial instead of Genkan class

    base_url = 'https://reaperscans.com'
    search_url = base_url + '/comics'
    most_populars_url = base_url + '/home'
    manga_url = base_url + '/comics/{0}'
    chapter_url = base_url + '/comics/{0}/{1}'
    image_url = base_url + '{0}'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from urllib.parse import urlparse

from bs4 import BeautifulSoup
import requests

from komikku.servers.multi.madara import Madara2


class Toonily(Madara2):
    id = 'toonily'
    name = 'Toonily'
    lang = 'en'
    is_nsfw = True

    date_format = '%b %-d, %y'
    medium = None
    series_name = 'serie'

    base_url = 'https://toonily.com'
    logo_url = base_url + '/wp-content/uploads/2024/01/cropped-toonfavicon_color_changed-32x32.png'
    chapter_url = base_url + '/' + series_name + '/{0}/{1}/'

    results_selector = '.manga'
    result_name_slug_selector = '.post-title a'
    result_cover_selector = '.item-thumb img'

    def is_long_strip(self, _manga_data):
        return True

    def search(self, term, nsfw, orderby=None):
        params = {}
        if orderby is None:
            params.update({
                's': term or '',
                'post_type': 'wp-manga',
                'author': '',
                'artist': '',
                'release': '',
            })
        elif orderby == 'populars':
            params['m_orderby'] = 'views'
        elif orderby == 'latest':
            params['m_orderby'] = 'latest'

        cookie = requests.cookies.create_cookie(
            name='toonily-mature',
            value=str(int(nsfw)),
            domain=urlparse(self.base_url).netloc,
            path='/',
            expires=None,
        )
        self.session.cookies.set_cookie(cookie)

        r = self.session_get(f'{self.base_url}/webtoons/', params=params)
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for element in soup.select('.page-item-detail'):
            thumb_element = element.select_one('.item-thumb')
            if not thumb_element.a:
                continue

            last_chapter_element = element.select_one('.chapter-item .chapter')
            if cover_element := thumb_element.a.img:
                cover = cover_element.get('src')
            else:
                cover = None

            results.append(dict(
                slug=thumb_element.a.get('href').split('/')[-2],
                name=thumb_element.a.get('title').strip(),
                cover=cover,
                last_chapter=last_chapter_element.text.strip() if last_chapter_element else None,
            ))

        return results



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from bs4 import BeautifulSoup

from komikku.servers.multi.wpcomics import WPComics
from komikku.utils import get_buffer_mime_type


class Jmanga(WPComics):
    id = 'jmanga'
    name = 'JManga'
    lang = 'ja'
    is_nsfw = True
    status = 'disabled'  # 2024/05 new site https://mangaraw.onl

    base_url = 'https://jmanga.vip'
    search_url = base_url + '/search/manga'
    latest_updates_url = base_url + '/search/manga?status=-1'
    most_populars_url = base_url + '/search/manga?status=-1&sort=11'
    manga_url = base_url + '/manga/{0}'
    chapter_url = base_url + '/manga/{0}/{1}'

    details_name_selector = '.breadcrumb li:last-child span'
    details_cover_selector = '#item-detail .col-image img'
    details_status_selector = '#item-detail .list-info .status p:last-child'
    details_authors_selector = '#item-detail .list-info .author p:last-child'
    details_genres_selector = '#item-detail .list-info .kind p:last-child a'
    details_synopsis_selector = '#item-detail .detail-content'
    results_link_selector = 'h3 a'
    results_cover_img_selector = '.box_img a img'
    results_last_chapter_link_selector = '.comic-item .chapter a'
    results_last_chapter_lastest_updates_link_selector = '.hlb-list li a'

    def get_latest_updates(self):
        """
        Returns latest updates

        Contrary to the standard WPComics theme, there is no latest updates dedicated page, so we use search instead.
        """
        r = self.session.get(
            self.latest_updates_url,
            headers={
                'Referer': self.search_url,
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for element in soup.select('.items .item'):
            a_element = element.select_one(self.results_link_selector)
            img_element = element.select_one(self.results_cover_img_selector)
            last_a_element = element.select_one(self.results_last_chapter_link_selector)

            results.append(dict(
                name=a_element.text.strip(),
                slug=a_element.get('href').split('/')[-1],
                cover=img_element.get('data-original'),
                last_chapter=last_a_element.text.replace('Issue', '').strip() if last_a_element else None,
            ))

        return results



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.madara import Madara2


class Starboundscans(Madara2):
    id = 'starboundscans'
    name = 'Starbound Scans'
    lang = 'fr'
    status = 'disabled'

    base_url = 'https://starboundscans.com'
    logo_url = base_url + '/wp-content/uploads/2025/07/cropped-g_logo-1-32x32.png'
    chapter_url = base_url + '/manga/{0}/{1}/'

    details_authors_selector = '.info-item:-soup-contains("Auteur") a'
    details_scanlators_selector = '.info-item:-soup-contains("Team") a'
    details_genres_selector = '.info-item:-soup-contains("Genres") .genre-tag'
    details_status_selector = '.info-item:-soup-contains("Statut") .info-value'
    details_synopsis_selector = '.card-body .black-orion-article-content p'

    chapters_selector = '.chapters-list .chapter-content'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from gettext import gettext as _
import time

import requests

from komikku.consts import DOWNLOAD_MAX_DELAY
from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type
from komikku.utils import get_response_elapsed
from komikku.utils import is_number

SEARCH_RESULTS_PAGES = 2
SEARCH_RESULTS_PAGE_LIMIT = 98


class Weebdex(Server):
    id = 'weebdex'
    name = 'WeebDex'
    lang = 'en'
    lang_code = 'en'
    status = 'disabled'  # EoS 04/07/2026

    is_nsfw = True

    base_url = 'https://weebdex.org'
    logo_url = base_url + '/favicon.ico'

    manga_url = base_url + '/title/{0}'
    chapter_url = base_url + '/chapter/{0}'
    cover_url = base_url + '/covers/{0}/{1}.256.webp'

    api_url = 'https://api.weebdex.org'
    api_search_url = api_url + '/manga'
    api_manga_url = api_url + '/manga/{0}'
    api_chapters_url = api_url + '/manga/{0}/chapters'
    api_chapter_url = api_url + '/chapter/{0}'

    filters = [
        {
            'key': 'ratings',
            'type': 'select',
            'name': _('Rating'),
            'description': _('Filter by Content Ratings'),
            'value_type': 'multiple',
            'options': [
                {'key': 'safe', 'name': _('Safe'), 'default': True},
                {'key': 'suggestive', 'name': _('Suggestive'), 'default': True},
                {'key': 'erotica', 'name': _('Erotica'), 'default': False},
                {'key': 'pornographic', 'name': _('Pornographic'), 'default': False},
            ]
        },
        {
            'key': 'statuses',
            'type': 'select',
            'name': _('Status'),
            'description': _('Filter by Statuses'),
            'value_type': 'multiple',
            'options': [
                {'key': 'ongoing', 'name': _('Ongoing'), 'default': False},
                {'key': 'complete', 'name': _('Completed'), 'default': False},
                {'key': 'hiatus', 'name': _('Hiatus'), 'default': False},
                {'key': 'cancelled', 'name': _('Canceled'), 'default': False},
            ]
        },
        {
            'key': 'demographics',
            'type': 'select',
            'name': _('Demographic'),
            'description': _('Filter by Publication Demographics'),
            'value_type': 'multiple',
            'options': [
                {'key': 'shounen', 'name': _('Shounen'), 'default': False},
                {'key': 'shoujo', 'name': _('Shoujo'), 'default': False},
                {'key': 'josei', 'name': _('Josei'), 'default': False},
                {'key': 'seinen', 'name': _('Seinen'), 'default': False},
            ]
        },
    ]

    headers = {
        'User-Agent': USER_AGENT,
    }

    long_strip_genres = [
        'Long Strip',
    ]

    params = [
        {
            'key': 'data_saver',
            'type': 'checkbox',
            'name': _('Data Saver Mode'),
            'description': _('Fetch lower quality images to save bandwidth'),
            'default': False,
        },
    ]

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers = self.headers

    def get_manga_data(self, initial_data):
        """
        Returns manga data from API

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Slug is missing in initial data'

        r = self.session_get(self.api_manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        resp_data = r.json()

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],
            genres=[],
            status=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
            cover=None,
        ))

        data['name'] = resp_data['title']
        if cover := resp_data['relationships'].get('cover'):
            data['cover'] = self.cover_url.format(resp_data['id'], cover['id'])

        # Details
        if resp_data['status'] == 'completed':
            data['status'] = 'complete'
        elif resp_data['status'] == 'ongoing':
            data['status'] = 'ongoing'
        elif resp_data['status'] == 'hiatus':
            data['status'] = 'hiatus'
        elif resp_data['status'] == 'cancelled':
            data['status'] = 'suspended'

        if tags := resp_data['relationships'].get('tags'):
            for tag in tags:
                data['genres'].append(tag['name'])

        if rating := resp_data.get('content_rating'):
            if rating != 'safe':
                data['genres'].append(rating.capitalize())

        if demographic := resp_data.get('demographic'):
            data['genres'].append(demographic.capitalize())

        if authors := resp_data['relationships'].get('authors'):
            for author in authors:
                data['authors'].append(author['name'])

        if artists := resp_data['relationships'].get('artists'):
            for artist in artists:
                if artist['name'] in data['authors']:
                    continue
                data['authors'].append(artist['name'])

        if groups := resp_data['relationships'].get('available_groups'):
            for group in groups:
                data['scanlators'].append(group['name'])

        if synopsis := resp_data.get('description'):
            data['synopsis'] = synopsis

        # Chapters
        data['chapters'] = self.get_manga_chapters_data(data['slug'])

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data from API

        Currently, only pages are expected.
        """
        r = self.session_get(
            self.api_chapter_url.format(chapter_slug),
            headers={
                'Referer': self.chapter_url.format(chapter_slug),
            }
        )
        if r.status_code != 200:
            return None

        resp_data = r.json()

        data_saver = self.get_param('data_saver')
        if data_saver and resp_data.get('data_optimized'):
            pages = resp_data['data_optimized']
        else:
            pages = resp_data['data']

        data = dict(
            pages=[],
        )
        node_url = resp_data['node']
        for index, page in enumerate(pages, start=1):
            data['pages'].append(dict(
                slug=None,
                image=f'{node_url}/data/{chapter_slug}/{page["name"]}',
                index=index,
            ))

        return data

    def get_manga_chapters_data(self, slug, page=1, chapters=None):
        if chapters is None:
            chapters = []

        r = self.session_get(
            self.api_chapters_url.format(slug),
            params={
                'tlang': self.lang_code,
                'page': page,
                'limit': 250,
            },
            headers={
                'Referer': self.manga_url.format(slug),
            }
        )
        if r.status_code != 200:
            return None

        rtime = get_response_elapsed(r)
        resp_data = r.json()

        if resp_data['total'] == 0:
            return chapters

        for chapter in resp_data['data']:
            title = []
            if num_volume := chapter.get('volume'):
                title.append(f'Vol. {num_volume}')
            if num := chapter.get('chapter'):
                title.append(f'Ch. {num}')
            if chapter.get('title'):
                title.append(chapter['title'])
            elif num_volume is None and num is None:
                title.append(_('Oneshot'))

            groups = chapter['relationships'].get('groups')

            chapters.append(dict(
                slug=chapter['id'],
                title=' '.join(title),
                num=num if is_number(num) else None,
                num_volume=num_volume if is_number(num_volume) else None,
                scanlators=[group['name'] for group in groups] if groups else None,
                date=convert_date_string(chapter['published_at'].split('T')[0], '%Y-%m-%d'),
            ))

        if page * 250 < resp_data['total']:
            if rtime:
                time.sleep(min(rtime * 4, DOWNLOAD_MAX_DELAY))

            self.get_manga_chapters_data(slug, page=page+1, chapters=chapters)

        return list(reversed(chapters))

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(page['image'])
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name='{0:03d}.{1}'.format(page['index'], mime_type.split('/')[-1]),
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_manga_list(self, term=None, demographics=None, ratings=None, statuses=None, orderby=None):
        def get_page(page):
            params = {
                'hasChapters': 'true',
                'availableTranslatedLang': [self.lang_code],
                'limit': SEARCH_RESULTS_PAGE_LIMIT,
            }
            if term:
                params['title'] = term

            if demographics:
                params['demographic'] = demographics
            if ratings:
                params['contentRating'] = ratings
            if statuses:
                params['status'] = statuses

            if orderby == 'popular':
                params['sort'] = 'views'
            elif term:
                params['sort'] = 'relevance'

            r = self.session_get(
                self.api_search_url,
                params=params,
                headers={
                    'Referer': f'{self.base_url}/',
                }
            )
            if r.status_code != 200:
                return [], False, None

            resp_data = r.json()

            more = page * SEARCH_RESULTS_PAGE_LIMIT < resp_data['total'] and page < SEARCH_RESULTS_PAGES

            return resp_data.get('data', []), more, get_response_elapsed(r)

        results = []
        delay = None
        more = True
        page = 1
        while more:
            if delay:
                time.sleep(delay)

            items, more, rtime = get_page(page)
            for item in items:
                cover = item['relationships'].get('cover')

                results.append(dict(
                    slug=item['id'],
                    name=item['title'],
                    cover=self.cover_url.format(item['id'], cover['id']) if cover else None,
                    last_chapter=item.get('last_chapter'),
                ))

            delay = min(rtime * 4, DOWNLOAD_MAX_DELAY) if rtime else None
            page += 1

        return results

    def get_latest_updates(self, demographics=None, ratings=None, statuses=None):
        return self.get_manga_list(demographics=demographics, ratings=ratings, statuses=statuses, orderby='latest')

    def get_most_populars(self, demographics=None, ratings=None, statuses=None):
        return self.get_manga_list(demographics=demographics, ratings=ratings, statuses=statuses, orderby='popular')

    def search(self, term, demographics=None, ratings=None, statuses=None):
        return self.get_manga_list(term=term, demographics=demographics, ratings=ratings, statuses=statuses)



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.manga_stream import MangaStream


class Lelmanga(MangaStream):
    id = 'lelmanga'
    name = 'Lelmanga'
    lang = 'fr'

    slug_position = -1

    base_url = 'https://www.lelmanga.com'
    logo_url = 'https://i3.wp.com/www.lelmanga.com/wp-content/uploads/2021/03/cropped-lelmanga-favicon-32x32.png'
    manga_url = base_url + '/manga/{0}'

    authors_selector = '.tsinfo .imptdt:-soup-contains("Artiste") i, .tsinfo .imptdt:-soup-contains("Auteur") i'
    genres_selector = '.info-desc .mgen a'
    scanlators_selector = '.tsinfo .imptdt:-soup-contains("Sérialisation") i'
    status_selector = '.tsinfo .imptdt:-soup-contains("Status") i'
    synopsis_selector = '[itemprop="description"]'

    def get_manga_data(self, initial_data):
        data = MangaStream.get_manga_data(self, initial_data)
        if data is None:
            return None

        def convert(s):
            # Fix bad encoding
            try:
                s = s.replace('“', '"').replace('”', '"').replace('’', "'").replace('…', '...').encode('iso-8859-1').decode()
            except Exception:
                pass

            return s

        for key in ('authors', 'genres', 'scanlators', 'synopsis'):
            if isinstance(data[key], list):
                for index, value in enumerate(data[key]):
                    data[key][index] = convert(value)
            elif isinstance(data[key], str):
                data[key] = convert(data[key])

        return data



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.madara import Madara


class Mangascantrad(Madara):
    id = 'mangascantrad'
    name = 'Manga-Scantrad'
    lang = 'fr'
    is_nsfw = True

    has_cf = True

    date_format = None

    base_url = 'https://manga-scantrad.io'
    chapters_url = base_url + '/manga/{0}/ajax/chapters/'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import base64
from gettext import gettext as _
import logging

from bs4 import BeautifulSoup

from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type
from komikku.utils import is_number
from komikku.webview import CompleteChallenge

logger = logging.getLogger(__name__)


class Raijinscan(Server):
    id = 'raijinscan'
    name = 'Raijin Scan'
    lang = 'fr'
    has_cf = True

    base_url = 'https://raijin-scans.fr'
    logo_url = base_url + '/wp-content/uploads/2025/05/cropped-logopp-32x32.png'
    search_url = base_url + '/'
    manga_url = base_url + '/manga/{0}/'
    chapter_url = base_url + '/manga/{0}/{1}/'
    bypass_cf_url = base_url + '/manga/nano-machine-1/'

    filters = [
        {
            'key': 'statuses',
            'type': 'select',
            'name': _('Status'),
            'description': _('Filter by Statuses'),
            'value_type': 'multiple',
            'options': [
                {'key': 'on-going', 'name': _('Ongoing'), 'default': False},
                {'key': 'end', 'name': _('Completed'), 'default': False},
            ],
        },
        {
            'key': 'types',
            'type': 'select',
            'name': _('Type'),
            'description': _('Filter by Types'),
            'value_type': 'multiple',
            'options': [
                {'key': 'manga', 'name': _('Manga'), 'default': False},
                {'key': 'manhwa', 'name': _('Manhwa'), 'default': False},
                {'key': 'manhua', 'name': _('Manhua'), 'default': False},
            ],
        },
    ]
    long_strip_genres = ['Manhwa', 'Webtoon']

    def __init__(self):
        self.session = None

    @CompleteChallenge()
    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Manga slug is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],  # not available
            genres=[],
            status=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
            cover=None,
        ))

        data['name'] = soup.select_one('.serie-title').text.strip()
        data['cover'] = soup.select_one('img.cover').get('src')

        # Details
        data['genres'] = [element.text.strip() for element in soup.select('.genre-list .grx div')]
        if element := soup.select_one('.stat-item .stat-details:-soup-contains("Type") .manga'):
            type = element.text.strip()
            if type not in data['genres']:
                data['genres'].append(type)

        if element := soup.select_one('.stat-item .stat-details:-soup-contains("État du titre") .manga'):
            status = element.text.strip()
            if status == 'En cours':
                data['status'] = 'ongoing'
            elif status == 'Terminé':
                data['status'] = 'complete'

        if element := soup.select_one('.stat-item .stat-details:-soup-contains("Auteur") .stat-value'):
            data['authors'].append(element.text.strip())
        if element := soup.select_one('.stat-item .stat-details:-soup-contains("Artiste") .stat-value'):
            artist = element.text.strip()
            if artist not in data['authors']:
                data['authors'].append(artist)

        # Synopsis
        data['synopsis'] = soup.select_one('.description').text.strip()

        # Chapters
        for element in reversed(soup.select('ul li.item')):
            if 'premium-chapter' in element.get('class'):
                continue
            a_element = element.select_one('a')
            date_element = a_element.select_one('.chapter-meta')
            slug = a_element.get('href').split('/')[-1]

            data['chapters'].append(dict(
                slug=slug,
                title=a_element.get('title'),
                num=slug if is_number(slug) else None,
                date=convert_date_string(date_element.text.strip(), languages=[self.lang]) if date_element else None,
            ))

        return data

    @CompleteChallenge()
    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Currently, only pages are expected.
        """
        r = self.session_get(self.chapter_url.format(manga_slug, chapter_slug))
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = dict(
            pages=[],
        )
        for element in soup.select('.protected-image-data'):
            data['pages'].append(dict(
                image=base64.b64decode(element.get('data-src')).decode('utf-8'),
                slug=None,
            ))

        return data

    @CompleteChallenge()
    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(
            page['image'],
            headers={
                'Referer': self.chapter_url.format(manga_slug, chapter_slug),
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=page['image'].split('/')[-1],
        )

    @CompleteChallenge()
    def get_latest_updates(self, statuses=None, types=None):
        """
        Returns recent mangas
        """
        return self.search(None, statuses=statuses, types=types, orderby='recently_added')

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    @CompleteChallenge()
    def get_most_populars(self, statuses=None, types=None):
        """
        Returns most viewed mangas
        """
        return self.search(None, statuses=statuses, types=types, orderby='most_viewed')

    @CompleteChallenge()
    def search(self, term, statuses=None, types=None, orderby=None):
        params = {
            'post_type': 'wp-manga',
            's': term or '',
        }
        if statuses:
            params['status[]'] = statuses
        if types:
            params['type[]'] = types
        if orderby:
            params['sort'] = orderby

        r = self.session_get(self.base_url, params=params)
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for element in soup.select('.unit'):
            a_element = element.select_one('.info a')
            img_element = element.select_one('.poster > div > img')
            last_chapter_element = element.select_one('.info ul li:first-child .ch-num')

            results.append({
                'slug': a_element.get('href').split('/')[-2],
                'name': a_element.text.strip(),
                'cover': img_element.get('src'),
                'last_chapter': last_chapter_element.text.strip().split()[-1],
            })

        return results



# SPDX-FileCopyrightText: 2020-2024 GrownNed
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: GrownNed <grownned@gmail.com>

from datetime import datetime

import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.utils import get_buffer_mime_type

SERVER_NAME = 'Desu'

headers = {
    'User-Agent': USER_AGENT,
}


class Desu(Server):
    id = 'desu'
    name = SERVER_NAME
    lang = 'ru'
    is_nsfw = True

    base_url = 'https://desu.uno'
    logo_url = base_url + '/styles/favicons/favicon-32x32.png?v=1'
    api_url = base_url + '/manga/api/'
    api_latest_updates_url = base_url + '/manga/api?limit=50&order=updated'
    api_most_populars_url = base_url + '/manga/api?limit=50&order=popular'
    api_manga_url = base_url + '/manga/api/{0}'
    api_chapter_url = base_url + '/manga/api/{0}/chapter/{1}'
    api_search_url = base_url + '/manga/api?limit=50&search={0}'

    manga_title_css_selector = 'h1 > span.name'

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers = headers

    def get_manga_data(self, initial_data):
        """
        Returns manga data from API

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Slug is missing in initial data'

        r = self.session_get(self.api_manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        resp_data = r.json()['response']

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],
            genres=[],
            status=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
        ))

        data['name'] = resp_data['russian']
        data['url'] = resp_data['url']
        data['cover'] = resp_data['image']['original']

        if resp_data.get('translators'):
            data['scanlators'] = [t['name'] for t in resp_data['translators']]
        data['genres'] = [genre['russian'] for genre in resp_data['genres']]
        if resp_data['status'] == 'ongoing':
            data['status'] = 'ongoing'
        elif resp_data['status'] == 'released':
            data['status'] = 'complete'
        data['synopsis'] = resp_data['description']

        for chapter in reversed(resp_data['chapters']['list']):
            title = '#{0}'.format(chapter['ch'])
            if chapter['title']:
                title = '{0} - {1}'.format(title, chapter['title'])

            data['chapters'].append(dict(
                slug=chapter['id'],
                title=title,
                num=chapter['ch'],
                date=datetime.fromtimestamp(chapter['date']).date(),
            ))

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data from API

        Currently, only pages are expected.
        """
        r = self.session_get(self.api_chapter_url.format(manga_slug, chapter_slug))
        if r.status_code != 200:
            return None

        resp_data = r.json()['response']

        data = dict(
            pages=[],
        )
        for page in resp_data['pages']['list']:
            data['pages'].append(dict(
                slug=None,
                image=page['img'],
            ))

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(page['image'], headers={'Referer': self.base_url})
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=page['image'].split('/')[-1],
        )

    @staticmethod
    def get_manga_url(slug, url):
        """
        Returns manga absolute URL
        """
        return url

    def get_latest_updates(self):
        """
        Returns latest updated mangas
        """
        return self.search('', orderby='updated')

    def get_most_populars(self):
        """
        Returns most popular mangas (bayesian rating)
        """
        return self.search('', orderby='popular')

    def search(self, term, orderby=None):
        params = dict(
            limit=50,
        )
        if orderby is not None:
            params['order'] = orderby
        else:
            params['search'] = term

        r = self.session_get(self.api_url, params=params, headers={'Referer': self.base_url})
        if r.status_code != 200:
            return None

        resp_data = r.json()['response']

        return [dict(
            slug=item['id'],
            name=item['russian'],
            cover=item['image']['preview'],
            last_chapter=item['chapters']['updated']['ch'],
        ) for item in resp_data]



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

# flake8: noqa: F401

from .database import backup_db
from .database import create_db_connection
from .database import delete_rows
from .database import init_db
from .database import insert_rows
from .database import update_rows
from .database.categories import Category
from .database.categories import CategoryVirtual
from .database.downloads import Download
from .database.mangas import Chapter
from .database.mangas import Manga

from .settings import Settings



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from gettext import gettext as _
import threading

from gi.repository import Adw
from gi.repository import GLib
from gi.repository import Gtk

from komikku.models import Category
from komikku.models import CategoryVirtual
from komikku.models import create_db_connection
from komikku.models import delete_rows
from komikku.models import insert_rows
from komikku.models import Settings


class CategoriesList:
    edit_mode = False  # mode to edit categories (of a manga selection) in bulk

    def __init__(self, library):
        self.library = library
        self.listbox = self.library.categories_listbox
        self.stack = self.library.categories_stack
        self.edit_mode_buttonbox = self.library.categories_edit_mode_buttonbox

        self.listbox.connect('row-activated', self.on_category_activated)
        self.library.categories_edit_mode_ok_button.connect('clicked', self.on_edit_mode_ok_button_clicked)
        self.library.categories_edit_mode_cancel_button.connect('clicked', self.on_edit_mode_cancel_button_clicked)

        self.set_togglebutton_accent()

    def clear(self):
        row = self.listbox.get_first_child()
        while row:
            next_row = row.get_next_sibling()
            self.listbox.remove(row)
            row = next_row

    def on_category_activated(self, _listbox, row):
        if self.edit_mode:
            return

        Settings.get_default().selected_category = row.category.id if isinstance(row.category, Category) else row.category

        self.set_togglebutton_accent()

        self.listbox.unselect_all()
        self.listbox.select_row(row)

        self.library.update_title()
        self.library.flowbox.invalidate_filter()

    def on_edit_mode_cancel_button_clicked(self, _button):
        self.library.overlaysplitview.set_show_sidebar(False)

    def on_edit_mode_ok_button_clicked(self, _button):
        def run():
            insert_data = []
            delete_data = []

            # List of selected manga
            manga_ids = []
            for thumbnail in self.library.flowbox.get_selected_children():
                manga_ids.append(thumbnail.manga.id)

            for row in self.listbox:
                if row.get_activatable_widget().get_active():
                    if Settings.get_default().selected_category == row.category.id:
                        # No insert, we are sure that category is already associated with all selected manga
                        continue

                    associated_manga_ids = row.category.mangas
                    for manga_id in manga_ids:
                        if manga_id in associated_manga_ids:
                            # No insert, category is already associated with this manga
                            continue

                        insert_data.append(dict(
                            manga_id=manga_id,
                            category_id=row.category.id,
                        ))
                elif Settings.get_default().selected_category == row.category.id:
                    for manga_id in manga_ids:
                        delete_data.append(dict(
                            manga_id=manga_id,
                            category_id=row.category.id,
                        ))

            db_conn = create_db_connection()
            with db_conn:
                if insert_data:
                    insert_rows(db_conn, 'categories_mangas_association', insert_data)
                if delete_data:
                    delete_rows(db_conn, 'categories_mangas_association', delete_data)

            GLib.idle_add(complete)

        def complete():
            self.library.window.activity_indicator.set_visible(False)

            # Leave library section mode and refresh library
            self.library.leave_selection_mode()
            self.library.populate()

        self.library.window.activity_indicator.set_visible(True)

        thread = threading.Thread(target=run)
        thread.daemon = True
        thread.start()

    def populate(self):
        db_conn = create_db_connection()
        categories = db_conn.execute('SELECT * FROM categories ORDER BY label ASC').fetchall()
        nb_categorized = db_conn.execute('SELECT count(*) FROM categories_mangas_association').fetchone()[0]

        if not categories and self.edit_mode:
            return

        self.clear()

        self.edit_mode_buttonbox.set_visible(self.edit_mode)

        if categories:
            self.stack.set_visible_child_name('list')

            items = ['all'] + categories
            if nb_categorized > 0:
                items += ['uncategorized']

            for item in items:
                nb = None
                if item == 'all':
                    if self.edit_mode:
                        continue

                    category = CategoryVirtual.ALL
                    label = _('All')
                elif item == 'uncategorized':
                    if self.edit_mode:
                        continue

                    category = CategoryVirtual.UNCATEGORIZED
                    label = _('Uncategorized')
                    nb = db_conn.execute('SELECT count(*) FROM mangas').fetchone()[0] - nb_categorized
                else:
                    category = Category.get(item['id'], db_conn=db_conn)
                    label = category.label
                    if not self.edit_mode:
                        nb = db_conn.execute(
                            'SELECT count(*) FROM categories_mangas_association WHERE category_id = ?', (item['id'],)
                        ).fetchone()[0]

                row = Adw.ActionRow(activatable=True)
                row.category = category
                row.set_title(label)

                if (isinstance(category, Category) and Settings.get_default().selected_category == category.id) or \
                        (isinstance(category, int) and Settings.get_default().selected_category == category):
                    self.listbox.select_row(row)

                if self.edit_mode:
                    switch = Gtk.Switch()
                    switch.set_active(Settings.get_default().selected_category == category.id)
                    switch.set_valign(Gtk.Align.CENTER)
                    row.set_activatable_widget(switch)
                    row.add_suffix(switch)
                elif nb:
                    nb_label = Gtk.Label(label=str(nb), valign=Gtk.Align.CENTER)
                    nb_label.add_css_class('badge')
                    nb_label.add_css_class('caption')
                    row.add_suffix(nb_label)

                self.listbox.append(row)
        else:
            Settings.get_default().selected_category = CategoryVirtual.ALL
            self.stack.set_visible_child_name('empty')

    def set_edit_mode(self, edit_mode):
        self.edit_mode = edit_mode

    def set_togglebutton_accent(self):
        if Settings.get_default().selected_category != CategoryVirtual.ALL:
            self.library.categories_togglebutton.add_css_class('accent')
        else:
            self.library.categories_togglebutton.remove_css_class('accent')



# SPDX-FileCopyrightText: 2021-2024 Lili Kurek
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Lili Kurek <lilikurek@proton.me>

from komikku.servers.multi.guya import Guya


class Dankefurslesen(Guya):
    id = 'dankefurslesen'
    name = 'Danke fürs Lesen'
    lang = 'en'
    is_nsfw = True

    base_url = 'https://danke.moe'
    logo_url = base_url + '/static/logo-mt-squared-small.png?v=a111ec8'
    manga_url = base_url + '/read/manga/{0}/'
    api_manga_url = base_url + '/api/series/{0}/'
    page_image_url = base_url + '/media/manga/{0}/chapters/{1}/{2}'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.manga_stream import MangaStream


class Ryujinmanga(MangaStream):
    id = 'ryujinmanga'
    name = 'Ryujinmanga'
    lang = 'es'
    status = 'disabled'

    base_url = 'https://www.ryujinmanga.com'

    authors_selector = '.tsinfo .imptdt:-soup-contains("Autor") i, .tsinfo .imptdt:-soup-contains("Artista") i'
    genres_selector = '.info-desc .mgen a'
    scanlators_selector = '.tsinfo .imptdt:-soup-contains("Publicado en") i'
    status_selector = '.tsinfo .imptdt:-soup-contains("Estado") i'
    synopsis_selector = '[itemprop="description"]'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from gettext import gettext as _
import threading

from gi.repository import Adw
from gi.repository import GLib
from gi.repository import Gtk

from komikku.consts import LOGO_SIZE


class TrackerRow(Adw.ActionRow):
    def __init__(self, window, tracker):
        self.window = window
        self.tracker = tracker

        super().__init__(title=self.tracker.name, activatable=False)

        if self.tracker.logo_path:
            logo = Gtk.Image()
            logo.set_pixel_size(LOGO_SIZE)
            logo.set_from_file(self.tracker.logo_path)
        else:
            logo = Adw.Avatar.new(LOGO_SIZE, self.tracker.name, True)

        self.add_prefix(logo)

        self.btn = Gtk.Button(valign=Gtk.Align.CENTER)

        granted, access_token_valid = self.tracker.is_granted()
        if granted and access_token_valid:
            self.active = True
            self.btn.set_label(_('Disconnect'))
            self.btn.set_css_classes(['destructive-action'])
        else:
            self.active = False

            if granted and not access_token_valid:
                self.set_subtitle(_('Connection is expired'))
                self.btn.set_label(_('Reconnect'))
                self.btn.set_css_classes(['suggested-action'])
            else:
                self.btn.set_label(_('Connect'))
                self.btn.set_css_classes([])

        self.btn.connect('clicked', self.on_btn_clicked)
        self.add_suffix(self.btn)

    def on_btn_clicked(self, _btn):
        group = None

        def open_dialog():
            nonlocal group

            group = Adw.PreferencesGroup()

            username_entry = Adw.EntryRow(title=_('Username'))
            username_entry.add_prefix(Gtk.Image.new_from_icon_name('avatar-default-symbolic'))
            group.add(username_entry)

            password_entry = Adw.PasswordEntryRow(title=_('Password'))
            password_entry.add_prefix(Gtk.Image.new_from_icon_name('dialog-password-symbolic'))
            group.add(password_entry)

            self.window.open_dialog(
                self.tracker.name,
                child=group,
                confirm_label=_('Connect'),
                confirm_callback=connect_dialog
            )

        def connect_dialog():
            username_entry = group.get_row(0)
            password_entry = group.get_row(1)
            success, error = self.tracker.request_access_token(username_entry.get_text(), password_entry.get_text())
            GLib.idle_add(connect_finish, success, error)

        def connect_webview():
            success, error = self.tracker.request_access_token()
            GLib.idle_add(connect_finish, success, error)

        def connect_finish(success, error):
            if success:
                self.active = True
                self.btn.set_label(_('Disconnect'))
                self.btn.set_css_classes(['destructive-action'])

            elif error == 'load_failed':
                self.window.preferences.add_toast(Adw.Toast.new(_('Failed to request client access')))

            elif error == 'locked':
                self.window.preferences.add_toast(Adw.Toast.new(_('Webview is currently in used. Please retry later.')))

            elif error == 'canceled':
                self.window.preferences.add_toast(Adw.Toast.new(error))

        def disconnect():
            self.tracker.data = None

            self.active = False
            self.btn.set_label(_('Connect'))
            self.btn.set_css_classes([])

        if not self.active:
            if self.tracker.authorize_url:
                thread = threading.Thread(target=connect_webview)
                thread.daemon = True
                thread.start()
            else:
                open_dialog()

        else:
            self.window.open_dialog(
                _('Disconnect from {}').format(self.tracker.name),
                confirm_label=_('Disconnect'),
                confirm_callback=disconnect,
                confirm_appearance=Adw.ResponseAppearance.DESTRUCTIVE
            )



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from gettext import gettext as _

from bs4 import BeautifulSoup
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type
from komikku.utils import ServerContent

# Conversion ISO_639-1 codes => server codes
LANGUAGES_CODES = {
    'cs': 'cs',
    'de': 'de',
    'en': 'en',
    'eo': 'eo',
    'es': 'es',
    'fa': 'fa',
    'fr': 'fr',
    'id': 'id',
    'it': 'it',
    'ja': 'ja',
    'ko': 'kr',  # diff
    'nb': 'no',  # diff
    'nl': 'nl',
    'pl': 'pl',
    'pt': 'pt',
    'ru': 'ru',
    'vi': 'vi',
    'zh_Hans': 'cn',  # diff
}

SERVER_NAME = 'Pepper & Carrot'


def convert_old_slug(slug):
    if slug != '':
        return slug

    # Old Peppercarrot slug was '' (empty string)
    return 'peppercarrot'


class Peppercarrot(Server):
    id = 'peppercarrot'
    name = SERVER_NAME
    lang = 'en'
    content = ServerContent(
        type=[_('Webcomic'), _('Self-publishing')],
        license='CC BY'
    )
    true_search = False

    base_url = 'https://www.peppercarrot.com'
    donate_url = base_url + '/en/support/index.html'
    logo_url = base_url + '/core/img/favicon.png'
    manga_url = base_url + '/{0}/webcomics/{1}.html'
    langs_url = base_url + '/0_sources/langs.json'
    chapters_url = base_url + '/0_sources/episodes-v1.json'
    chapter_url = base_url + '/{0}/{1}/{2}.html'
    image_url = base_url + '/0_sources/{0}/low-res/{1}_{2}'
    cover_url = base_url + '/core/img/thumb_{0}.jpg'

    genres = ['Fantasy', 'Magical girl', 'Coming-of-age']

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update({'User-Agent': USER_AGENT})

        self.donate_url = f'{self.base_url}/{LANGUAGES_CODES[self.lang]}/support/index.html'

    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content
        """
        r = self.session_get(
            self.manga_url.format(LANGUAGES_CODES[self.lang], convert_old_slug(initial_data['slug']))
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()
        data.update({
            'authors': ['David Revoy', ],
            'scanlators': [],
            'genres': self.genres,
            'status': 'ongoing',
            'synopsis': None,
            'chapters': [],
            'server_id': self.id,
            'cover': self.cover_url.format(convert_old_slug(data['slug'])),
        })

        if element := soup.select_one('.teaser span'):
            data['synopsis'] = element.text.strip()

        # Chapters
        if data['slug'] == 'miniFantasyTheater':
            # Chapters
            for element in reversed(soup.select('figure')):
                if 'notranslation' in element.get('class'):
                    continue

                a_element = element.select_one('figcaption a')
                date_element = element.select_one('figcaption span')

                data['chapters'].append({
                    'slug': a_element.get('href').split('/')[-1].split('.')[0],
                    'date': convert_date_string(date_element.text.strip().split()[-1][:-1], format='%Y-%m-%d'),
                    'title': a_element.text.strip(),
                })

        else:
            # Scanlators (translators here)
            r = self.session_get(self.langs_url)
            if r.ok:
                if lang_data := r.json().get(LANGUAGES_CODES[self.lang]):
                    data['scanlators'] = lang_data['translators']

            # Chapters
            r = self.session_get(self.chapters_url)
            if r.status_code != 200:
                return None

            chapters_data = r.json()

            for index, element in enumerate(reversed(soup.select('figure.thumbnail'))):
                if 'notranslation' in element.get('class'):
                    # Skipped not translated episodes
                    continue

                data['chapters'].append({
                    'slug': chapters_data[index]['name'],
                    'date': None,
                    'title': element.a.img.get('title').split('(')[0].strip(),
                })

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data
        """
        if manga_slug == 'miniFantasyTheater':
            r = self.session_get(self.chapter_url.format(LANGUAGES_CODES[self.lang], manga_slug, chapter_slug))
            if r.status_code != 200:
                return None

            soup = BeautifulSoup(r.text, 'lxml')

            return {
                'pages': [{
                    'slug': None,
                    'image': soup.select_one('.webcomic-page img').get('src'),
                }],
            }

        r = self.session_get(self.chapters_url)
        if r.status_code != 200:
            return None

        chapters_data = r.json()
        for chapter_data in chapters_data:
            if chapter_data['name'] == chapter_slug:
                pages = chapter_data['pages']
                break

        data = {
            'pages': [],
        }

        # Cover & Title pages are first
        data['pages'].append({
            'slug': pages.pop('cover'),
            'image': None,
        })
        data['pages'].append({
            'slug': pages.pop('title'),
            'image': None,
        })
        credits_slug = pages.pop('credits')

        # Sort pages
        pages = dict(sorted(pages.items(), key=lambda x: int(x[0])))

        for _key, page_name in pages.items():
            data['pages'].append({
                'slug': page_name,
                'image': None,
            })

        # Credits page at end
        data['pages'].append({
            'slug': credits_slug,
            'image': None,
        })

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        if page['image']:
            url = page['image']
        else:
            url = self.image_url.format(chapter_slug, LANGUAGES_CODES[self.lang], page['slug'])

        r = self.session_get(url)
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return {
            'buffer': r.content,
            'mime_type': mime_type,
            'name': url.split('/')[-1],
        }

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(LANGUAGES_CODES[self.lang], convert_old_slug(slug))

    def get_most_populars(self):
        return [
            {
                'slug': 'peppercarrot',
                'name': self.name,
                'cover': self.cover_url.format('peppercarrot'),
            },
            {
                'slug': 'miniFantasyTheater',
                'name': 'MiniFantasyTheater',
                'cover': self.cover_url.format('miniFantasyTheater'),
            }
        ]

    def search(self, term=None):
        # This server does not have a search
        # but a search method is needed for `Global Search` in `Explorer`
        # In order not to be offered in `Explorer`, class attribute `true_search` must be set to False

        results = []
        for item in self.get_most_populars():
            if term and term.lower() in item['name'].lower():
                results.append(item)

        return results


class Peppercarrot_cs(Peppercarrot):
    id = 'peppercarrot_cs'
    name = SERVER_NAME
    lang = 'cs'


class Peppercarrot_de(Peppercarrot):
    id = 'peppercarrot_de'
    name = SERVER_NAME
    lang = 'de'


class Peppercarrot_eo(Peppercarrot):
    id = 'peppercarrot_eo'
    name = SERVER_NAME
    lang = 'eo'


class Peppercarrot_es(Peppercarrot):
    id = 'peppercarrot_es'
    name = SERVER_NAME
    lang = 'es'


class Peppercarrot_fa(Peppercarrot):
    id = 'peppercarrot_fa'
    name = 'فلفل و هویج'
    lang = 'fa'


class Peppercarrot_fr(Peppercarrot):
    id = 'peppercarrot_fr'
    name = SERVER_NAME
    lang = 'fr'


class Peppercarrot_id(Peppercarrot):
    id = 'peppercarrot_id'
    name = SERVER_NAME
    lang = 'id'


class Peppercarrot_it(Peppercarrot):
    id = 'peppercarrot_it'
    name = SERVER_NAME
    lang = 'it'


class Peppercarrot_ja(Peppercarrot):
    id = 'peppercarrot_ja'
    name = SERVER_NAME
    lang = 'ja'


class Peppercarrot_ko(Peppercarrot):
    id = 'peppercarrot_ko'
    name = SERVER_NAME
    lang = 'ko'


class Peppercarrot_nb(Peppercarrot):
    id = 'peppercarrot_nb'
    name = SERVER_NAME
    lang = 'nb'


class Peppercarrot_nl(Peppercarrot):
    id = 'peppercarrot_nl'
    name = SERVER_NAME
    lang = 'nl'


class Peppercarrot_pl(Peppercarrot):
    id = 'peppercarrot_pl'
    name = SERVER_NAME
    lang = 'pl'


class Peppercarrot_pt(Peppercarrot):
    id = 'peppercarrot_pt'
    name = SERVER_NAME
    lang = 'pt'


class Peppercarrot_ru(Peppercarrot):
    id = 'peppercarrot_ru'
    name = SERVER_NAME
    lang = 'ru'


class Peppercarrot_vi(Peppercarrot):
    id = 'peppercarrot_vi'
    name = SERVER_NAME
    lang = 'vi'


class Peppercarrot_zh_hans(Peppercarrot):
    id = 'peppercarrot_zh_hans'
    name = SERVER_NAME
    lang = 'zh_Hans'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import base64
from bs4 import BeautifulSoup
import logging
import re
from urllib.parse import unquote

try:
    # This server requires JA3/TLS and HTTP2 fingerprints impersonation
    from curl_cffi import requests
except Exception:
    # Server will be disabled
    requests = None

from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type
from komikku.utils import is_number
from komikku.webview import CompleteChallenge

logger = logging.getLogger('komikku.servers.readcomiconline')


class Readcomiconline(Server):
    id = 'readcomiconline'
    name = 'Read Comic Online'
    lang = 'en'
    is_nsfw = True
    status = 'enabled' if requests is not None else 'disabled'

    has_cf = True
    has_captcha = True  # Custom captcha AreYouHuman2
    http_client = 'curl_cffi'

    base_url = 'https://readcomiconline.li'
    latest_updates_url = base_url + '/ComicList/LatestUpdate'
    most_populars_url = base_url + '/ComicList/MostPopular'
    search_url = base_url + '/AdvanceSearch'
    manga_url = base_url + '/Comic/{0}'
    chapter_url = base_url + '/Comic/{0}/{1}?s=&quality=hq#1'
    bypass_cf_url = base_url + '/Comic/Invincible/Issue-1#1'

    def __init__(self):
        self.session = None

    @CompleteChallenge()
    def get_manga_data(self, initial_data):
        """
        Returns comic data by scraping manga HTML page content

        Initial data should contain at least comic's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Slug is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['slug'], 1))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],  # not available
            genres=[],
            status=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
            cover=None,
        ))

        soup = BeautifulSoup(r.text, 'lxml')

        info_element = soup.select_one('#leftside .barContent')

        data['name'] = info_element.select_one('.bigChar').text.strip()
        cover_path = soup.select_one('#rightside img').get('src')
        if cover_path.startswith('http'):
            data['cover'] = cover_path
        else:
            data['cover'] = '{0}{1}'.format(self.base_url, cover_path)

        for p_element in info_element.select('p'):
            if not p_element.span:
                if not data['synopsis']:
                    data['synopsis'] = p_element.text.strip()
                continue

            span_element = p_element.span.extract()
            label = span_element.text.strip()

            if label.startswith('Genres'):
                data['genres'] = [a_element.text.strip() for a_element in p_element.select('a')]

            elif label.startswith(('Writer', 'Artist')):
                for a_element in p_element.select('a'):
                    value = a_element.text.strip()
                    if value not in data['authors']:
                        data['authors'].append(value)

            elif label.startswith('Status'):
                value = p_element.text.strip()
                if 'Completed' in value:
                    data['status'] = 'complete'
                elif 'Ongoing' in value:
                    data['status'] = 'ongoing'

        # Chapters (Issues)
        for tr_element in reversed(soup.select('.listing tr')):
            td_elements = tr_element.select('td')
            if not td_elements:
                continue

            slug = td_elements[0].a.get('href').split('?')[0].split('/')[-1]
            num = slug.split('-')[-1] if slug.startswith('Issue-') else None

            data['chapters'].append(dict(
                slug=slug,
                title=td_elements[0].a.text.strip(),
                num=num if is_number(num) else None,
                date=convert_date_string(td_elements[1].text.strip(), format='%m/%d/%Y'),
            ))

        return data

    @CompleteChallenge()
    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns comic chapter data
        """
        def decode_url(url, substitutions, base_url):
            # In `Scripts/rguard.min.js?v=1.5.8`
            if not url.startswith('https'):
                for substitution in substitutions:
                    url = url.replace(substitution[0], substitution[1])

                if '?' in url:
                    url, qs = url.split('?')
                else:
                    qs = None

                if '=s0' in url:
                    url = url.replace('=s0', '')
                    s = '=s0'
                elif '=s1600' in url:
                    url = url.replace('=s1600', '')
                    s = '=s1600'

                url = url[15:33] + url[50:]
                url = url[0:len(url) - 11] + url[len(url) - 2] + url[len(url) - 1]
                url = unquote(unquote(base64.b64decode(url)))
                url = 'https://2.bp.blogspot.com/' + url[0:13] + url[17:-2] + s
                if qs:
                    url += '?' + qs

            if base_url and base_url != '' and url.find('ip=') > 0:
                url = url.replace('https://2.bp.blogspot.com', base_url)

            return url

        r = self.session_get(self.chapter_url.format(manga_slug, chapter_slug))
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        var_rnd_name = None
        encoded_urls = []
        substitutions = []
        media_server = None
        for script_element in soup.select('script'):
            script = script_element.string
            if not script or ('var pth' not in script and 'replace(' not in script):
                continue

            for line in script.split('\n'):
                line = line.strip()

                # URLs (encoded) are stored in a variable with a random name
                if var_rnd_name is None:
                    # Search for a line of the form:
                    # var _1oXikvMxnz = '';
                    if matches := re.search(r"var\s+_([0-9a-zA-Z]+)\s+=\s+''", line):
                        var_rnd_name = '_' + matches.group(1)

                if var_rnd_name and line.startswith(var_rnd_name):
                    encoded_urls.append(line.split()[2][1:-2])

                elif 'replace(' in line:
                    if matches := re.search(r"replace\(/([a-zA-Z0-9_]+)/g, '([a-z])'\);", line):
                        substitutions.append((matches.group(1), matches.group(2)))

        data = dict(
            pages=[],
        )

        for index, url in enumerate(encoded_urls):
            data['pages'].append(dict(
                image=decode_url(url, substitutions, media_server),
                slug=None,
                index=index + 1,
            ))

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(
            page['image'] + '&t=10',
            headers={
                'Alt-Used': '2.bp.blogspot.com',
            }
        )

        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=f"{page['index']}.{mime_type.split('/')[1]}",
        )

    def get_manga_url(self, slug, url):
        """
        Returns comic absolute URL
        """
        return self.manga_url.format(slug)

    def get_manga_list(self, term=None, orderby=None):
        results = []

        if term:
            r = self.session_get(
                self.search_url,
                params=dict(
                    comicName=term,
                    ig='',
                    eg='',
                    status='',
                    pubDate='',
                ),
                headers={
                    'Referer': self.search_url,
                }
            )
        elif orderby == 'populars':
            r = self.session_get(self.most_populars_url)
        elif orderby == 'latest':
            r = self.session_get(self.latest_updates_url)
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        for a_element in soup.select('.item > a:first-child'):
            if not a_element.get('href'):
                continue

            cover = a_element.img.get('src')
            if not cover.startswith('http'):
                cover = self.base_url + cover

            results.append(dict(
                name=a_element.span.text.strip(),
                slug=a_element.get('href').split('/')[-1],
                cover=cover,
            ))

        return results

    @CompleteChallenge()
    def get_latest_updates(self):
        """
        Returns latest updates
        """
        return self.get_manga_list(orderby='latest')

    @CompleteChallenge()
    def get_most_populars(self):
        """
        Returns most popular comics
        """
        return self.get_manga_list(orderby='populars')

    @CompleteChallenge()
    def search(self, term):
        return self.get_manga_list(term=term)



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.madara import Madara


class Poseidonscans(Madara):
    id = 'poseidonscans'
    name = 'Poseidon Scans'
    lang = 'fr'
    is_nsfw = True
    status = 'disabled'  # 2025/05: New site with dynamic HTML (Next.js)

    has_cf = True

    date_format = None

    base_url = 'https://poseidonscans.fr'
    chapters_url = base_url + '/manga/{0}/ajax/chapters/'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import logging
import re

from bs4 import BeautifulSoup

from komikku.servers.multi.madara import Madara2
from komikku.utils import get_buffer_mime_type

logger = logging.getLogger(__name__)


class Mangacrab(Madara2):
    id = 'mangacrab'
    name = 'MangaCrab'
    lang = 'es'
    is_nsfw = True

    date_format = '%d/%m/%Y'
    series_name = 'series'

    base_url = 'https://mangacrab.org'
    logo_url = base_url + '/wp-content/uploads/2017/10/cropped-logo100-Personalizado-32x32.png'
    chapter_url = base_url + '/' + series_name + '/{0}/{1}/'

    details_name_selector = 'h1.post-title'
    details_status_selector = '.post-content_item:-soup-contains("Estado") .summary-content'

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Currently, only pages are expected.
        """
        r = self.session_get(
            self.chapter_url.format(manga_slug, chapter_slug),
            headers={
                'Referer': self.manga_url.format(manga_slug),
            }
        )
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        img_key = None
        for script_element in soup.find_all('script'):
            script = script_element.string
            if script and 'Img-Key' in script:
                re_key = r"'Img-Key': '(.*)'"
                if matches := re.search(re_key, script):
                    img_key = matches.group(1)
                    break

        if img_key is None:
            logger.info('Failed to retrieve Img-Key header value')
            return None

        data = dict(
            pages=[],
        )
        index = 1
        for img_element in soup.select('.page-break img.wp-manga-chapter-img'):
            image = None
            for attr, value in img_element.attrs.items():
                if isinstance(value, str) and 'encript.php' in value:
                    image = value
                    break
            if image is None:
                continue

            data['pages'].append(dict(
                slug=None,
                image=f'{self.base_url}/{image}',
                index=index,
                key=img_key,
            ))
            index += 1

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(
            page['image'],
            headers={
                'Accept': 'image/*',
                'Referer': self.chapter_url.format(manga_slug, chapter_slug),
                'Img-Key': page['key'],
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name='{0:03d}.{1}'.format(page['index'], mime_type.split('/')[-1]),
        )

    def search(self, term, nsfw, orderby=None):
        params = {
            'post_type': 'wp-manga',
            'type': 'manga',
        }
        if term:
            params['s'] = term

        if orderby == 'populars':
            params['m_orderby'] = 'views'
        elif orderby == 'latest':
            params['m_orderby'] = 'latest'

        r = self.session_get(f'{self.base_url}/', params=params)
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for element in soup.select('.search-lists .manga__item'):
            a_element = element.select_one('.post-title h2 a')
            thumb_element = element.select_one('.manga__thumb_item')
            nb_chapters_element = element.select_one('.manga-info .total')

            if cover_element := thumb_element.a.img or element.img:
                cover = cover_element.get('data-src')
                if not cover:
                    cover = cover_element.get('src')
            else:
                cover = None

            results.append(dict(
                slug=a_element.get('href').split('/')[-2],
                name=a_element.text.strip(),
                cover=cover,
                nb_chapters=nb_chapters_element.text.strip().split()[0] if nb_chapters_element else None,
            ))

        return results



# SPDX-FileCopyrightText: 2021-2024 Lili Kurek
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Lili Kurek <lilikurek@proton.me>

import datetime
import logging
import time

from bs4 import BeautifulSoup
import requests

from komikku.consts import DOWNLOAD_MAX_DELAY
from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type
from komikku.utils import get_response_elapsed

logger = logging.getLogger('komikku.servers.tapas')

# each page is 10 entries
SEARCH_RESULTS_PAGES = 10
CHAPTERS_PER_REQUEST = 20


class Tapas(Server):
    id = 'tapas'
    name = 'Tapas'
    lang = 'en'
    status = 'disabled'  # commercial platforms are no supported

    base_url = 'https://tapas.io'
    api_base_url = 'https://story-api.tapas.io'
    api_manga_list_url = api_base_url + '/cosmos/api/v1/landing/genre'
    search_url = base_url + '/search'
    manga_url = base_url + '/series/{0}'
    manga_info_url = base_url + '/series/{0}/info'
    chapters_url = base_url + '/series/{0}/episodes'
    chapter_url = base_url + '/episode/{0}'

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update({'user-agent': USER_AGENT})
            for name in ('birthDate', 'adjustedBirthDate'):
                cookie = requests.cookies.create_cookie(
                    name=name,
                    value='2001-01-01',
                    domain='tapas.io',
                    path='/',
                    expires=None,
                )
                self.session.cookies.set_cookie(cookie)

    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Manga slug is missing in initial data'

        r = self.session_get(self.manga_info_url.format(initial_data['slug']))
        if r is None:
            return None

        mime_type = get_buffer_mime_type(r.content)

        if r.status_code != 200 or mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],  # not available
            genres=[],
            status=None,  # not available
            synopsis=None,
            chapters=[],
            server_id=self.id,
            cover=None,
        ))

        data['name'] = soup.find(class_='section__top--simple').find(class_='title').text
        data['synopsis'] = soup.select_one('.description__body').text.strip()
        data['cover'] = soup.find(class_='section__top--simple').find(class_='thumb').img.get('src')

        for author_element in soup.find(class_='creator').find_all('a'):
            data['authors'].append(author_element.text)

        if element := soup.select_one('.tags'):
            # Community series
            for a_element in element.select('a'):
                data['genres'].append(a_element.text.strip().replace('#', ''))
        elif elements := soup.select('.detail-row__body--genre a'):
            for a_element in elements:
                data['genres'].append(a_element.text.strip())

        data['chapters'] = self.get_manga_chapters_data(initial_data['slug'])

        return data

    def get_manga_chapters_data(self, manga_slug):
        def get_page(page):
            r = self.session_get(
                self.chapters_url.format(manga_slug),
                params=dict(
                    page=page,
                    sort='OLDEST',
                    init_load=0,
                    max_limit=CHAPTERS_PER_REQUEST,
                    since=int(datetime.datetime.now().timestamp()) * 1000,
                    large='true',
                    last_access=0,
                )
            )
            if r.status_code != 200:
                return None

            resp_data = r.json()['data']
            more = resp_data['pagination']['has_next']

            return resp_data['episodes'], more, get_response_elapsed(r)

        chapters = []
        delay = None
        more = True
        page = 1
        while more:
            if delay:
                time.sleep(delay)

            episodes, more, rtime = get_page(page)
            for episode in episodes:
                if not episode['free'] or episode['must_pay'] or episode['scheduled']:
                    continue

                chapters.append(dict(
                    slug=str(episode['id']),  # slug nust be a string
                    title=episode['title'],
                    date=convert_date_string(episode['publish_date'].split('T')[0], format='%Y-%m-%d'),
                ))

            delay = min(rtime * 2, DOWNLOAD_MAX_DELAY) if rtime else None
            page += 1

        return chapters

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Currently, only pages are expected.
        """
        r = self.session_get(self.chapter_url.format(chapter_slug))
        if r is None:
            return None

        mime_type = get_buffer_mime_type(r.content)

        if r.status_code != 200 or mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = dict(
            slug=chapter_slug,
            title=soup.find(class_='title').text,
            pages=[],
            date=convert_date_string(soup.find(class_='date').text, format='%b %d, %Y'),
        )

        for page in soup.find(class_='js-episode-article').find_all('img'):
            data['pages'].append(dict(
                slug=None,
                image=page.get('data-src'),
            ))

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(page['image'])
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=page['image'].split('/')[-1],
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_manga_list(self, orderby):
        def get_page(page):
            r = self.session_get(
                self.api_manga_list_url,
                params=dict(
                    category_type='COMIC',
                    sort_option=orderby,
                    subtab_id=17,
                    page=page,
                    size=25,
                ),
                headers={
                    'Origin': self.base_url,
                    'Referer': f'{self.base_url}/',
                }
            )
            if r.status_code != 200:
                return None

            resp_data = r.json()

            more = not resp_data['meta']['pagination']['last'] and page < SEARCH_RESULTS_PAGES

            return resp_data['data']['items'], more, get_response_elapsed(r)

        delay = None
        more = True
        page = 0
        results = []
        while more:
            if delay:
                time.sleep(delay)

            items, more, rtime = get_page(page)
            for item in items:
                if item['bmType'] == 'WAIT_UNTIL_FREE':
                    continue

                if item['assetProperty'].get('backgroundCharacterImage'):
                    img_obj = item['assetProperty']['backgroundCharacterImage']
                else:
                    img_obj = item['assetProperty'].get('thumbnailImage')

                results.append(dict(
                    slug=item['seriesId'],
                    name=item['title'],
                    cover=img_obj['path'] + '.webp' if img_obj else None,
                ))

            delay = min(rtime * 2, DOWNLOAD_MAX_DELAY) if rtime else None
            page += 1

        return results

    def get_latest_updates(self):
        return self.get_manga_list(orderby='NEWEST_EPISODE')

    def get_most_populars(self):
        return self.get_manga_list(orderby='POPULAR')

    def search(self, term, page_number=1):
        r = self.session_get(
            self.search_url,
            params=dict(
                pageNumber=page_number,
                q=term,
                t='COMICS',
            ),
            headers={
                'Referer': self.search_url,
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for li_element in soup.select('.search-item-wrap'):
            a_element = li_element.select_one('.title-section a.link')

            if a_element.get('data-sale-type') not in ('EARLY_ACCESS', 'FREE', 'false', 'WAIT_OR_MUST_PAY'):
                continue

            results.append(dict(
                slug=a_element.get('data-series-id'),
                url=a_element.get('href'),
                name=a_element.text,
                cover=li_element.select_one('.item-thumb-wrap img').get('src'),
            ))

        if page_number == 1:
            if buttons := soup.select('a.paging__button--num'):
                last_page_number = int(buttons[-1].text)

                for page in range(2, min(SEARCH_RESULTS_PAGES + 1, last_page_number + 1)):
                    results += self.search(term, page)

        return results



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import datetime
import json
from urllib.parse import urlparse

from bs4 import BeautifulSoup
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type
from komikku.utils import is_number
from komikku.webview import eval_js


class Mangahere(Server):
    id = 'mangahere'
    name = 'MangaHere'
    lang = 'en'
    is_nsfw = True

    base_url = 'https://www.mangahere.cc'
    logo_url = base_url + '/favicon.ico'
    search_url = base_url + '/search'
    latest_updates_url = base_url + '/latest/'
    most_populars_url = base_url + '/ranking/'
    manga_url = base_url + '/manga/{0}/'
    chapter_url = base_url + '/manga/{0}/{1}/'
    page_url = base_url + '/manga/{0}/{1}/{2}.html'

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers = {
                'User-Agent': USER_AGENT,
            }
            cookie = requests.cookies.create_cookie(
                name='isAdult',
                value='1',
                domain=urlparse(self.base_url).netloc,
                path='/',
                expires=(datetime.datetime.now() + datetime.timedelta(days=1)).timestamp(),
            )
            self.session.cookies.set_cookie(cookie)

    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Slug is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        if r.url.startswith(self.search_url):
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],
            genres=[],
            status=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
            cover=None,
        ))

        data['name'] = soup.select_one('.detail-info-right-title-font').text.strip()
        data['cover'] = soup.select_one('.detail-info-cover-img').get('src')

        # Details
        status = soup.select_one('.detail-info-right-title-tip').text.strip().lower()
        data['status'] = 'ongoing' if status == 'ongoing' else 'complete'
        for a_element in soup.select('.detail-info-right-say a'):
            data['authors'].append(a_element.text.strip())
        for a_element in soup.select('.detail-info-right-tag-list a'):
            data['genres'].append(a_element.text.strip())

        # Synopsis
        data['synopsis'] = soup.select_one('.detail-info-right-content').text.strip()

        # Chapters
        for a_element in reversed(soup.select('.detail-main-list > li > a')):
            num = None
            num_volume = None
            slug = a_element.get('href').split('/')[3:-1]
            for part in slug:
                if part[0] == 'c':
                    # cXXX
                    num = part[1:]
                elif part[0] == 'v':
                    # vYY
                    num_volume = part[1:]

            data['chapters'].append(dict(
                slug='/'.join(slug),  # cXXX or vYY/cXXX
                title=a_element.get('title').replace(data['name'], '').strip(),
                num=num if is_number(num) else None,
                num_volume=num_volume if is_number(num_volume) else None,
                date=convert_date_string(a_element.select_one('.title2').text.strip(), format='%b %d,%Y'),
            ))

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Currently, only pages are expected.
        """
        r = self.session_get(self.chapter_url.format(manga_slug, chapter_slug))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        for script_element in soup.find_all('script'):
            script = script_element.string
            if script is None:
                continue

            if script.strip().startswith('eval(function(p,a,c,k,e,d)'):
                js_code = 'pasglop = ' + script.strip()[5:-1] + '; pasglop'

            elif 'var chapterid' in script:
                # Get chapter ID, required with page by page reader
                for line in script.split(';'):
                    if not line.strip().startswith('var imagecount'):
                        continue

                    nb_pages = int(line.strip().split('=')[-1].strip())
                    break

        data = dict(
            pages=[],
        )

        if len(soup.select('.reader-main > div')) == 0:
            # Webtoon reader
            res = eval_js(js_code)
            # We obtain something like this
            # var newImgs=['url1','url2',...];blabla...;
            pages_urls = json.loads(res[12:].split(';')[0].replace("'", '"'))

            for url in pages_urls:
                data['pages'].append(dict(
                    image=f'https:{url}',  # noqa: E231
                ))
        else:
            # Page by page reader
            for index in range(1, nb_pages):
                data['pages'].append(dict(
                    index=index,
                ))

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        if page.get('index'):
            index = page['index']
            r = self.session_get(self.page_url.format(manga_slug, chapter_slug, index))
            if r.status_code != 200:
                return None

            mime_type = get_buffer_mime_type(r.content)
            if mime_type != 'text/html':
                return None

            soup = BeautifulSoup(r.text, 'lxml')

            # We need a secret key and the chapter ID
            for script_element in soup.find_all('script'):
                script = script_element.string
                if script is None:
                    continue

                if script.strip().startswith('eval(function(p,a,c,k,e,d)'):
                    js_code = 'pasglop = ' + script.strip()[5:-1] + '; pasglop'

                elif 'var chapterid' in script:
                    # Get chapter ID
                    for line in script.split(';'):
                        if line.strip().startswith('var chapterid'):
                            cid = line.strip().split('=')[-1].strip()
                            break

            res = eval_js(js_code)
            # We obtain something like this
            # var guidkey=''+'3'+'a'+'e'+'d'+'6'+'a'+'a'+'6'+'1'+'0'+'4'+'e'+'7'+'2'+'3'+'e';blabla...;
            key = res[13:].split(';')[0][:-1].replace("'+'", '')

            r = self.session_get(
                self.chapter_url.format(manga_slug, chapter_slug) + f'chapterfun.ashx?cid={cid}&page={index}&key={key}',
                headers={
                    'Referer': self.page_url.format(manga_slug, chapter_slug, index),
                    'X-Requested-With': 'XMLHttpRequest',
                }
            )

            js_code = 'pasglop = ' + r.text[5:-2] + '; pasglop'
            res = eval_js(js_code)
            # We obtain something like this
            # function dm5imagefun(){var pix="//the_base_url";var pvalue=["path_image","path_next_image"];blabla...;
            base_url = res[23:-1].split(';')[0][9:-1]
            images = json.loads(res.split(';')[1][11:])

            url = f'https:{base_url}{images[0]}'  # noqa: E231
        else:
            url = page['image']

        r = self.session_get(url, headers={'Referer': self.base_url})
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=url.split('/')[-1].split('_')[-1],
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_latest_updates(self):
        """
        Returns Latest Updates
        """
        return self.search(None, orderby='latest')

    def get_most_populars(self):
        """
        Returns Most Popular
        """
        return self.search(None, orderby='populars')

    def search(self, term, orderby=None):
        if term:
            r = self.session_get(self.search_url, params=dict(title=term))
        elif orderby == 'latest':
            r = self.session_get(self.latest_updates_url)
        elif orderby == 'populars':
            r = self.session_get(self.most_populars_url)
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for li_element in soup.select('.line-list > div> ul > li'):
            results.append(dict(
                slug=li_element.a.get('href').split('/')[-2],
                name=li_element.a.get('title').strip(),
                cover=li_element.a.img.get('src'),
            ))

        return results



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.madara import Madara2


class Mangasorigines(Madara2):
    id = 'mangasorigines'
    name = 'Mangas Origines'
    lang = 'fr'
    is_nsfw = True

    series_name = 'oeuvre'

    base_url = 'https://mangas-origines.fr'
    logo_url = base_url + '/wp-content/uploads/2023/07/cropped-favmo3-32x32.png'
    chapters_url = base_url + '/' + series_name + '/{0}/ajax/chapters/?t=1'

    details_cover_selector = 'picture > img'
    chapters_date_selector = '.timediff'

    images_src_attr = 'data-src'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from gettext import gettext as _
from bs4 import BeautifulSoup
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import get_soup_element_inner_text
from komikku.utils import get_buffer_mime_type
from komikku.utils import ServerContent

# Conversion ISO_639-1 codes => server codes
LANGUAGES_CODES = dict(
    en='english',
    es='espanol',
    fr='francais',
    ru='ru',
    zh_Hans='zh',  # diff
)


class Dbnewhope(Server):
    id = 'dbnewhope'
    name = 'DB New Hope'
    lang = 'fr'
    content = ServerContent(
        type=[_('Doujinshi'), _('Fan Webcomic')]
    )
    true_search = False

    base_url = 'https://dbnewhope.com'
    donate_url = base_url + '/#patreon'
    logo_url = base_url + '/assets/favicon.ico'
    manga_url = base_url + '/{0}/chapitres.php'
    page_url = base_url + '/{0}/{1}/01.php?page={2}'
    cover_url = base_url + '/francais/1/fe07eee1f0268775c646e844b05bff2d.png?ezimgfmt=rs%3Adevice%2Frscb1-1'

    synopsis = """
Au lendemain de sa victoire écrasante contre Gohan lors du Cell Game, Cell s'est éclipsé de la Terre, laissant Krilin indemne mais marqué par le souvenir de sa propre misère. Sept années se sont écoulées, le monde s'est reconstruit, mais la menace persiste dans l'ombre. Krilin, se charge de former le dernier guerrier survivant de la planète, en prévision d'un éventuel retour du monstre créé par le machiavélique Dr. Gero.

Le dernier espoir de la Z-Team se révèle être un individu au caractère bien trempé, à la fierté débordante. Entre les leçons de combat et les entraînements intensifs, une relation complexe se tisse, mêlant respect, apprentissage et compréhension mutuelle. Les pages de ce manga se dévoilent comme un nouvel épisode palpitant de l'univers épique de DBZ, où le passé douloureux et les défis du présent convergent vers un avenir incertain.

Dans ce récit inédit, découvrez une aventure trépidante où les liens entre maître et apprenti se forgent au fil des combats, où l'héritage des guerriers Z se perpétue face à une menace insidieuse. Sera-t-il possible pour le dernier défenseur de la Terre d'apprivoiser la force intérieure de son protégé et de prévenir le retour redouté de Cell? Plongez-vous dans cette épopée captivante, où l'esprit de DBZ demeure plus vibrant que jamais.
    """

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update({'User-Agent': USER_AGENT})

    @property
    def lang_code(self):
        return LANGUAGES_CODES[self.lang]

    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content
        """
        r = self.session_get(self.manga_url.format(self.lang_code))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()
        data.update(dict(
            authors=['Burichan ブリちゃん', ],
            scanlators=[],
            genres=['Shōnen', 'Dōjinshi'],
            status='ongoing',
            synopsis=self.synopsis.strip(),
            chapters=[],
            server_id=self.id,
            cover=self.cover_url,
        ))

        # Chapters
        for element in soup.select('.card'):
            slug = element.a.get('href').split('/')[0]
            data['chapters'].append(dict(
                slug=slug,
                title=get_soup_element_inner_text(element.select_one('.card-title'), recursive=False),
                num=slug,
                date=None,
            ))

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga data by scraping manga HTML page content
        """
        r = self.session_get(self.page_url.format(self.lang_code, chapter_slug, 1))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = dict(
            pages=[],
        )
        for element in soup.select('.nav select option'):
            data['pages'].append(dict(
                slug=element.get('value').split('=')[-1],
                image=None,
            ))

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(self.page_url.format(self.lang_code, chapter_slug, page['slug']))
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        url = None
        if img_element := soup.select_one('.image-wrapper img'):
            url = img_element.get('data-ezsrc')
            if not url:
                url = img_element.get('src')

        if not url:
            return None

        r = self.session_get(
            self.base_url + url,
            headers={
                'Referer': self.page_url.format(self.lang_code, chapter_slug, page['slug']),
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name='{0}.{1}'.format(page['slug'], mime_type.split('/')[-1]),
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(self.lang_code)

    def get_most_populars(self):
        return [dict(
            slug=f'{self.lang_code}readdbnh',
            name=self.name,
            cover=self.cover_url,
        )]

    def search(self, term=None):
        # This server does not have a true search
        # but a search method is needed for `Global Search` in `Explorer`
        # In order not to be offered in `Explorer`, class attribute `true_search` must be set to False

        results = []
        for item in self.get_most_populars():
            if term and term.lower() in item['name'].lower():
                results.append(item)

        return results


class Dbnewhope_en(Dbnewhope):
    id = 'dbnewhope_en'
    lang = 'en'

    base_url = 'https://dbnewhope.com'
    manga_url = base_url + '/{0}/chapters.php'

    synopsis = """
The day after his crushing victory over Gohan in the Cell Game, Cell vanished from Earth, leaving Krilin unharmed but scarred by the memory of his own misery. Seven years have passed, the world has been rebuilt, but the threat persists in the shadows. Krilin takes it upon himself to train the planet's last surviving warrior, in anticipation of the possible return of the monster created by the Machiavellian Dr. Gero.

The last hope of the Z-Team turns out to be a strong-willed individual with an overflowing sense of pride. Between combat lessons and intensive training, a complex relationship develops, combining respect, learning and mutual understanding. The pages of this manga reveal themselves as a thrilling new episode in the epic DBZ universe, where the painful past and the challenges of the present converge towards an uncertain future.

In this all-new story, discover a thrilling adventure where the bonds between master and apprentice are forged in battle, where the legacy of the Z warriors lives on in the face of an insidious threat. Can Earth's last defender tame his protégé's inner strength and prevent Cell's dreaded return? Immerse yourself in this captivating epic, where the spirit of DBZ remains as vibrant as ever.
    """


class Dbnewhope_es(Dbnewhope):
    id = 'dbnewhope_es'
    lang = 'es'

    base_url = 'https://dbnewhope.com'
    manga_url = base_url + '/{0}/capitulos.php'

    synopsis = """
Al día siguiente de su aplastante victoria sobre Gohan en el Cell Game, Célula desapareció de la Tierra, dejando a Krilin ileso pero marcado por el recuerdo de su propia miseria. Han pasado siete años, el mundo ha sido reconstruido, pero la amenaza persiste en las sombras. Krilin se encarga de entrenar al último guerrero superviviente del planeta, en previsión del posible regreso del monstruo creado por el maquiavélico Dr. Gero.

La última esperanza del Equipo Z resulta ser un individuo de carácter fuerte y orgullo desbordante. Entre lecciones de combate y entrenamientos intensivos, se desarrolla una compleja relación que combina respeto, aprendizaje y comprensión mutua. Las páginas de este manga se revelan como un nuevo y emocionante episodio del épico universo DBZ, donde el doloroso pasado y los retos del presente convergen hacia un futuro incierto.

En esta historia totalmente nueva, descubre una emocionante aventura en la que los lazos entre maestro y aprendiz se forjan en la batalla, donde el legado de los guerreros Z sigue vivo frente a una insidiosa amenaza. ¿Será capaz el último defensor de la Tierra de domar la fuerza interior de su protegido y evitar el temido regreso de Célula? Sumérgete en esta cautivadora epopeya, donde el espíritu de DBZ sigue tan vivo como siempre.
    """


class Dbnewhope_ru(Dbnewhope):
    id = 'dbnewhope_ru'
    lang = 'ru'
    status = 'disabled'

    synopsis = """
На следующий день после сокрушительной победы над Гоханом в Игре Клеток, Клетка исчез с Земли, оставив Крилина невредимым, но с шрамами от воспоминаний о собственных страданиях. Прошло семь лет, мир был восстановлен, но угроза продолжает оставаться в тени. Крилин берется обучить последнего выжившего воина планеты в ожидании возможного возвращения монстра, созданного макиавеллистским доктором Геро.

Последняя надежда команды Z-Team оказывается волевым человеком с переполненным чувством гордости. Между боевыми уроками и интенсивными тренировками складываются сложные отношения, сочетающие в себе уважение, обучение и взаимопонимание. Страницы этой манги открываются как новый захватывающий эпизод в эпической вселенной DBZ, где болезненное прошлое и вызовы настоящего сходятся в неопределенном будущем.

В этой новой истории вас ждет захватывающее приключение, где узы между мастером и учеником скрепляются в бою, где наследие воинов Z продолжает жить перед лицом коварной угрозы. Сможет ли последний защитник Земли укротить внутреннюю силу своего протеже и предотвратить страшное возвращение Клетки? Погрузитесь в эту захватывающую эпопею, где дух DBZ остается таким же ярким, как и прежде.
    """


class Dbnewhope_zh_hans(Dbnewhope):
    id = 'dbnewhope_zh_hans'
    lang = 'zh_Hans'
    status = 'disabled'

    synopsis = """
细胞在 "细胞游戏 "中战胜悟饭后的第二天，就从地球上消失了，克里林毫发无损，但却留下了痛苦的回忆。七年过去了，世界已经重建，但威胁依然存在。克里林开始训练地球上最后一名幸存的战士，以防那个由马基雅维利斯-吉罗博士创造的怪物可能卷土重来。

Z 战队最后的希望竟然是一个意志坚强、自尊心极强的人。在战斗课程和强化训练之间，他们建立起了一种复杂的关系，将尊重、学习和相互理解融为一体。在这部漫画中，我们将看到史诗般的 DBZ 宇宙中惊心动魄的新篇章，痛苦的过去和现在的挑战交织在一起，走向不确定的未来。

在这个全新的故事中，您将发现一场惊心动魄的冒险，师徒之间的羁绊在战斗中铸就，Z 战士的传奇在阴险的威胁面前得以延续。地球最后的捍卫者能否驯服他的徒弟的内在力量，阻止可怕的细胞卷土重来？请沉浸在这部引人入胜的史诗中，DBZ 的精神将一如既往地充满活力。
    """



# SPDX-FileCopyrightText: 2021-2024 Lili Kurek
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Lili Kurek <lilikurek@proton.me>

import komikku.servers.multi.guya


class Guya(komikku.servers.multi.guya.Guya):
    id = 'guya'
    name = 'Guya.moe'
    lang = 'en'

    base_url = 'https://guya.cubari.moe'
    logo_url = base_url + '/static/logo_small.png?v=a6fe86b'
    manga_url = base_url + '/read/manga/{0}/'
    api_manga_url = base_url + '/api/series/{0}/'
    page_image_url = base_url + '/media/manga/{0}/chapters/{1}/{2}'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

# Genkan CMS

# Supported servers
# Edelgarde Scans [EN] (disabled)
# Hatigarm Scans [EN] (disabled)
# Leviatan Scans [EN/ES] (disabled)
# One Shot Scans [EN] (disabled)
# Reaper Scans [EN] (disabled)
# The Nonames Scans [EN] (disabled)

import json

from bs4 import BeautifulSoup
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type


class Genkan(Server):
    base_url: str
    search_url: str
    most_populars_url: str
    manga_url: str
    chapter_url: str
    image_url: str

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update({'user-agent': USER_AGENT})

    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Manga slug is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[self.name, ],
            genres=[],
            status=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
            cover=None,
        ))

        data['name'] = soup.find_all('h5')[0].text.strip()
        cover_url = soup.find('div', class_='media-comic-card').a.get('style').split('(')[-1][:-1]
        if cover_url.startswith('http'):
            data['cover'] = cover_url
        else:
            data['cover'] = self.image_url.format(cover_url)

        # Details
        data['synopsis'] = soup.find('div', class_='col-lg-9').contents[2].strip()

        # Chapters
        elements = soup.find('div', class_='list list-row row').find_all('div', class_='list-item')
        for element in reversed(elements):
            a_elements = element.find_all('a')

            slug = '/'.join(a_elements[0].get('href').split('/')[-2:])
            title = '#{0} - {1}'.format(element.span.text.strip(), a_elements[0].text.strip())
            date = a_elements[1].text.strip()

            data['chapters'].append(dict(
                slug=slug,
                date=convert_date_string(date),
                title=title,
            ))

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Currently, only pages are expected.
        """
        r = self.session_get(self.chapter_url.format(manga_slug, chapter_slug))
        if r is None:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if r.status_code != 200 or mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = dict(
            pages=[],
        )
        for script_element in soup.find_all('script'):
            script = script_element.string
            if script is None or not script.strip().startswith('window.disqusName'):
                continue

            for line in script.split(';'):
                line = line.strip()
                if not line.startswith('window.chapterPages'):
                    continue

                images = json.loads(line.split('=')[1].strip())
                for image in images:
                    data['pages'].append(dict(
                        slug=None,
                        image=image,
                    ))
                break
            break

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        if page['image'].startswith('http'):
            image_url = page['image']
        else:
            image_url = self.image_url.format(page['image'])

        r = self.session_get(image_url, headers={'Referer': self.chapter_url.format(manga_slug, chapter_slug)})
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=page['image'].split('/')[-1],
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_most_populars(self):
        """
        Returns new and/or recommended manga
        """
        r = self.session_get(self.most_populars_url)

        if r.status_code == 200:
            soup = BeautifulSoup(r.text, 'lxml')

            results = []
            for a_element in soup.find_all('a', class_='list-title ajax'):
                result = dict(
                    slug=a_element.get('href').split('/')[-1],
                    name=a_element.text.strip(),
                )
                if result not in results:
                    results.append(result)

            return results

        return None

    def search(self, term):
        r = self.session_get(self.search_url.format(term))

        if r.status_code == 200:
            soup = BeautifulSoup(r.text, 'lxml')

            results = []
            for a_element in soup.find_all('a', class_='list-title ajax'):
                results.append(dict(
                    slug=a_element.get('href').split('/')[-1],
                    name=a_element.text.strip(),
                ))

            return results

        return None


class GenkanInitial(Genkan):
    """
    The initial version of the CMS doesn't provide search
    """

    def search(self, term):
        r = self.session_get(self.search_url)

        if r.status_code == 200:
            soup = BeautifulSoup(r.text, 'lxml')

            results = []
            for a_element in soup.find_all('a', class_='list-title ajax'):
                name = a_element.text.strip()
                if term.lower() not in name.lower():
                    continue

                results.append(dict(
                    slug=a_element.get('href').split('/')[-1],
                    name=name,
                ))

            return results

        return None
